import typing
from dataclasses import dataclass
from typing import Any, Callable, Dict, Final, List, Optional, Tuple, TypeVar, final

from eth_abi import grammar
from eth_abi.grammar import ABIType, TupleType
from eth_typing import TypeStr


_T = TypeVar("_T")


Normalizer = Callable[[TypeStr, Any], Tuple[TypeStr, Any]]
MapperKey = Tuple[Tuple[Normalizer, ...], Tuple[TypeStr, ...]]
DataTreeFunc = Callable[[TypeStr, Any], Tuple[TypeStr, Any]]


# cdef typing
Iterable: Final = typing.Iterable
Mapping: Final = typing.Mapping


parse: Final[Callable[[TypeStr], ABIType]] = grammar.parse


@final
class Formatter:
    def __init__(
        self,
        normalizers: Tuple[Normalizer, ...],
        types: Tuple[Optional[TypeStr], ...],
    ):
        self.normalizers: Final = tuple(get_data_tree_map(n) for n in normalizers)
        self.types: Final[Tuple[Optional[ABIType], ...]] = tuple(
            parse(t) if isinstance(t, str) else None for t in types
        )

    def __call__(self, data: Any) -> List[Any]:
        # 1. Decorating the data tree with types
        # TODO: vendor abi_sub_tree, its v wasteful
        data = map(abi_sub_tree, self.types, data)
        # 2. Recursively mapping each of the normalizers to the data
        for n in self.normalizers:
            data = n(data)
        # 3. Stripping the types back out of the tree
        return list(strip_abi_types(data))


_formatters: Final[Dict[MapperKey, Formatter]] = {}


# web3py builds the pipeline every call, we cache it here instead
def get_formatter(
    normalizers: Tuple[Normalizer, ...],
    types: Tuple[TypeStr, ...],
) -> Formatter:
    """Returns a cached :class:`~Formatter` for the unique combination of normalizers and types."""
    mapper = _formatters.get((normalizers, types))
    if mapper is None:
        mapper = _formatters[(normalizers, types)] = Formatter(normalizers, types)
    return mapper


_data_tree_maps: Final[Dict[DataTreeFunc, "map_to_typed_data"]] = {}


def get_data_tree_map(
    func: DataTreeFunc,
) -> "map_to_typed_data":
    f = _data_tree_maps.get(func)
    if f is None:
        f = _data_tree_maps[func] = map_to_typed_data(func)
    return f


@final
class map_to_typed_data:
    def __init__(self, func: DataTreeFunc) -> None:
        self.func: Final = func

    def __call__(self, elements: Any) -> Any:
        datatype = type(elements)
        if datatype is list or datatype is map:
            return [self(obj) for obj in elements]
        elif datatype is tuple:
            return tuple(self(obj) for obj in elements)
        elif isinstance(elements, Mapping):
            return datatype((key, self(val)) for key, val in elements.items())  # type: ignore [call-arg]
        elif not isinstance(elements, (bytes, str, bytearray)) and isinstance(elements, Iterable):
            return datatype(map(self, elements))
        elif isinstance(elements, ABITypedData) and elements.abi_type is not None:
            return ABITypedData(*self.func(*elements))  # type: ignore [misc]
        else:
            return elements


def strip_abi_types(data: Any) -> Any:
    datatype = type(data)
    if datatype is list or datatype is map:
        return [strip_abi_types(obj) for obj in data]
    elif isinstance(data, tuple):
        return tuple(strip_abi_types(obj) for obj in data)
    elif isinstance(data, Mapping):
        return datatype((key, strip_abi_types(val)) for key, val in data.items())  # type: ignore [call-arg]
    elif not isinstance(data, (bytes, str, bytearray)) and isinstance(data, Iterable):
        return datatype(map(strip_abi_types, data))
    elif isinstance(data, ABITypedData):
        return data.data
    else:
        return data


# vendored from web3.py so we can compile it


@final
@dataclass(frozen=True)
class ABITypedData:
    """
    This class marks data as having a certain ABI-type.

    >>> a1 = ABITypedData('address', addr1)
    >>> a2 = ABITypedData('address', addr2)
    >>> addrs = ABITypedData('address[]', [a1, a2])

    You can access the fields using tuple() interface, or with
    attributes:

    >>> assert a1.abi_type == a1[0]
    >>> assert a1.data == a1[1]

    Unlike a typical `namedtuple`, you initialize with a single
    positional argument that is iterable, to match the init
    interface of all other relevant collections.
    """

    # NOTE this class was changed to a dataclass so it compiles to C better

    abi_type: Optional[TypeStr]
    data: Any


def abi_sub_tree(abi_type: Optional[ABIType], data_value: Any) -> ABITypedData:
    # TODO: specialize this function, possibly with functools.singledispatch

    if abi_type is None:
        return ABITypedData(None, data_value)

    # In the two special cases below, we rebuild the given data structures with
    # annotated items
    if abi_type.is_array:
        # If type is array, determine item type and annotate all
        # items in iterable with that type
        item_type = abi_type.item_type
        value_to_annotate = [abi_sub_tree(item_type, item_value) for item_value in data_value]
    elif isinstance(abi_type, TupleType):
        # Otherwise, if type is tuple, determine component types and annotate
        # tuple components in iterable respectively with those types
        value_to_annotate = type(data_value)(
            abi_sub_tree(comp_type, comp_value)
            for comp_type, comp_value in zip(abi_type.components, data_value)
        )
    else:
        value_to_annotate = data_value

    return ABITypedData(abi_type.to_type_str(), value_to_annotate)



from asyncio import (
    CancelledError,
    Future,
    Task,
    TimeoutError,
    create_task,
    current_task,
    get_running_loop,
    shield,
    sleep,
    wait_for,
)
from collections import defaultdict
from concurrent.futures.process import BrokenProcessPool
from itertools import chain, filterfalse, groupby
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Dict,
    Final,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    final,
)
from weakref import ProxyType
from weakref import proxy as weak_proxy

import a_sync
import eth_retry
from a_sync import AsyncProcessPoolExecutor, PruningThreadPoolExecutor, igather
from a_sync.asyncio import sleep0 as yield_to_loop
from a_sync.functools import cached_property_unsafe as cached_property
from aiohttp.client_exceptions import ClientResponseError
from eth_typing import ChecksumAddress
from eth_utils.toolz import concat
from hexbytes import HexBytes
from requests.exceptions import ReadTimeout
from web3.exceptions import ContractLogicError
from web3.types import RPCEndpoint, RPCResponse
from web3.types import RPCError as _RPCError

from dank_mids import ENVIRONMENT_VARIABLES as ENVS
from dank_mids import _debugging, constants, stats
from dank_mids._demo_mode import demo_logger
from dank_mids._exceptions import (
    BadResponse,
    BatchResponseSortError,
    ChainstackRateLimitError,
    DankMidsClientResponseError,
    DankMidsInternalError,
    EmptyBatch,
    ExceedsMaxBatchSize,
    ExecutionReverted,
    OutOfGas,
    PayloadTooLarge,
    RateLimitError,
    internal_err_types,
)
from dank_mids._logging import DEBUG, getLogger
from dank_mids.exceptions import GarbageCollectionError
from dank_mids.helpers import DebuggableFuture, _codec, _session, batch_size, gatherish
from dank_mids.helpers._codec import (
    JSONRPCBatchResponse,
    MulticallChunk,
    RawResponse,
    mcall_decode,
    mcall_encode,
)
from dank_mids.helpers._errors import (
    INDIVIDUAL_CALL_REVERT_STRINGS,
    error_logger,
    error_logger_debug,
    error_logger_log_debug,
    gas_logger_debug,
    is_call_revert,
    is_revert_bytes,
    log_internal_error,
    needs_full_request_spec,
    revert_logger,
    revert_logger_log_debug,
    timeout_logger_debug,
    timeout_logger_warning,
)
from dank_mids.helpers._gather import first_completed
from dank_mids.helpers._helpers import set_done
from dank_mids.helpers._lock import AlertingRLock
from dank_mids.helpers._multicall import MulticallContract
from dank_mids.helpers._session import rate_limit_inactive
from dank_mids.helpers._weaklist import WeakList
from dank_mids.helpers.method import get_len as get_len_for_method
from dank_mids.helpers.method import should_batch as should_batch_method
from dank_mids.types import (
    BatchId,
    BlockId,
    JsonrpcParams,
    PartialRequest,
    PartialResponse,
    Request,
    Response,
)

if TYPE_CHECKING:
    from dank_mids._batch import DankBatch
    from dank_mids.controller import DankMiddlewareController


TIMEOUT_SECONDS_SMALL: Final = 30
TIMEOUT_SECONDS_BIG: Final = float(ENVS.STUCK_CALL_TIMEOUT)  # type: ignore [arg-type]

logger: Final = getLogger(__name__)

_Response = TypeVar(
    "_Response", Response, List[Response], RPCResponse, List[RPCResponse], RawResponse
)


@final
class RPCError(_RPCError, total=False):
    dankmids_added_context: Dict[str, Any]


_super_init: Final = a_sync.Event.__init__
_super_set: Final = a_sync.Event.set


class _RequestEvent(a_sync.Event):
    def __init__(self, owner: "_RequestBase[_Response]") -> None:
        _super_init(self, debug_daemon_interval=300)
        if self.debug_logs_enabled:
            self._owner: ProxyType[_RequestBase[_Response]] = weak_proxy(owner)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} object at {hex(id(self))} [{'set' if self.is_set() else 'unset'}, waiter:{self._owner}>"

    def set(self) -> None:
        # Make sure we wake up the _RequestEvent's event loop if its in another thread
        if self._loop is get_running_loop():
            _super_set(self)
        else:
            self._loop.call_soon_threadsafe(_super_set, self)

    # default if no debug logs enabled
    _owner = "[not displayed...]"


class _RequestBase(Generic[_Response]):
    _fut: DebuggableFuture
    _batch: Optional["_Batch"] = None

    __slots__ = "controller", "uid", "_fut", "__weakref__"

    def __init__(self, controller: "DankMiddlewareController", uid: Optional[str] = None) -> None:
        self.controller: "DankMiddlewareController" = controller
        """The DankMiddlewareController that created this request."""

        self.uid = controller.call_uid.next if uid is None else uid
        """The unique id for this request."""

        self._fut = DebuggableFuture(self, controller._loop)

    def __await__(self) -> Generator[Any, None, _Response]:
        return self.get_response().__await__()

    # Abstract methods to be implemented by subclasses

    def __len__(self) -> int:
        raise NotImplementedError

    async def get_response(self) -> _Response:
        raise NotImplementedError


### Single requests:
_REVERT_EXC_TYPES: Final = ContractLogicError, ExecutionReverted

_request_base_init: Final = _RequestBase.__init__


class RPCRequest(_RequestBase[RPCResponse]):
    should_batch: bool = True
    """`True` if this request should be batched with others into a jsonrpc batch request, `False` if it should be sent as an individual request."""

    _tiny_batches: bool = False
    """`True` if the max jsonrpc batch size is 50 or lower."""

    _debug_logs_enabled: bool = False
    """`True` if debug logging is currently enabled."""

    method: Final[RPCEndpoint]
    params: Final[Any]
    raw: Final[bool]
    _fut: Final[DebuggableFuture]
    _daemon: Final[Optional["Task[None]"]]
    __dict__: Final[Dict[str, Any]]

    __slots__ = "method", "params", "raw", "_daemon", "__dict__"

    def __init__(
        self,
        controller: "DankMiddlewareController",
        method: RPCEndpoint,
        params: Any,
        uid: Optional[str] = None,
    ) -> None:  # sourcery skip: hoist-statement-from-if
        _request_base_init(self, controller, uid)
        if method[-4:] == "_raw":
            self.method = method[:-4]
            """The rpc method for this request."""
            self.raw = True
        else:
            self.method = method
            """The rpc method for this request."""
            self.raw = False

        self.params = params
        """The parameters to send with this request, if any."""

        if not should_batch_method(method):
            self.should_batch = False

        if logger.isEnabledFor(DEBUG):
            self._debug_logs_enabled = True

        if controller.max_jsonrpc_batch_size <= 50:
            self._tiny_batches = True

        if self.should_batch:
            with controller.pools_closed_lock:
                if isinstance(self, eth_call) and self.multicall_compatible:
                    controller.pending_eth_calls[self.block].append(self)
                elif self.method.startswith(("trace", "debug")):
                    raise NotImplementedError("we should not get here", self)
                else:
                    controller._pending_rpc_calls_append(self)
        _demo_logger_info("added to queue (cid: %s)", self.uid)
        if _logger_is_enabled_for(DEBUG):
            self._daemon = create_task(self._debug_daemon(), name="RPCRequest debug daemon")

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, __o: object) -> bool:
        return (self.uid == __o.uid) if type(__o) is type(self) else False

    def __len__(self) -> int:
        # NOTE: We dont need to consider response size for each method for very small batch sizes since the requests/responses will never get too large
        return 1 if self._tiny_batches else get_len_for_method(self.method)

    def __repr__(self) -> str:
        batch = self._batch
        batch_info = "" if batch is None else f" batch={batch}"
        return f"<{self.__class__.__name__} uid={self.uid} method={self.method} params={self.params}{batch_info}>"

    def __del__(self) -> None:
        fut = self._fut
        if not fut.done() and not fut._loop.is_closed():
            try:
                fut.set_exception(
                    GarbageCollectionError(
                        f"{self} was garbage collected before finishing.\n"
                        "This exception exists to help debug an issue inside of dank mids. Please show it to Bob."
                    )
                )
            except RuntimeError as e:
                if str(e) != "no running event loop":
                    raise
            else:
                # mark exception as retrieved, if its really relevant to the user it will be raised by their waiter
                fut.exception()

    @property
    def request(self) -> Union[Request, PartialRequest]:
        return self.controller.request_type(method=self.method, params=self.params, id=self.uid)

    async def get_response(self) -> RPCResponse:  # type: ignore [override]
        if not self.should_batch:
            if self._debug_logs_enabled:
                _log_debug("bypassed dank batching, method is %s", self.method)
            return await self.get_response_unbatched()

        current_batch = self._batch
        if current_batch is None:
            # NOTE: We want to force the event loop to make one full _run_once call before we execute.
            await yield_to_loop()

        elif current_batch._awaited is False:
            # NOTE: If current_batch is not None, that means we filled a batch. Let's await it now so we can send something to the node.
            await first_completed(current_batch._task, self._fut)

        fut = self._fut

        if self._batch is None:
            try:
                batch_task = create_task(
                    self.controller.execute_batch(), name="batch task execute_batch"
                )
                # If this timeout fails, we go nuclear and destroy the batch.
                # Any calls that already succeeded will have already completed on the client side.
                # Any calls that have not yet completed with results will be recreated, rebatched (potentially bringing better results?), and retried
                await wait_for(shield(batch_task), timeout=TIMEOUT_SECONDS_BIG)
            except TimeoutError:
                _log_debug(
                    "%s got stuck awaiting its batch, we're creating a new one",
                    self,
                )

                duplicate = self.create_duplicate()

                # don't start counting for the timeout while we still have a queue of requests to send
                await rate_limit_inactive(self.controller.endpoint)

                duplicate_task = create_task(
                    duplicate.get_response(), name="duplicate task get_response"
                )
                done, pending = await first_completed(batch_task, fut, duplicate_task)
                for d in done:
                    if d in (batch_task, fut):
                        # we'll get and decode the value below
                        continue
                    elif fut.done():
                        # if our fut is also done we can just return the result now
                        return d.result()
                    try:
                        # but if it isn't, we need to set the result before returning
                        # so our garbage collection code doesn't trigger
                        result = d.result()
                    except Exception as e:
                        fut.set_exception(e)
                    else:
                        fut.set_result(result)
                        return result

                # We did not get our result from the duplicate task, if it
                # ends up with an exception we don't need to know about it
                duplicate_task._Future__log_traceback = False

        try:
            if fut.done():
                response = fut.result()
            else:
                try:
                    response = await wait_for(shield(fut), timeout=TIMEOUT_SECONDS_BIG)  # type: ignore [arg-type]
                except CancelledError:
                    fut.cancel()
                    raise
                except TimeoutError:
                    _log_debug(
                        "%s got stuck waiting for its fut, we're creating a new one",
                        self,
                    )

                    duplicate = self.create_duplicate()

                    # don't start counting for the timeout while we still have a queue of requests to send
                    await rate_limit_inactive(self.controller.endpoint)

                    duplicate_task = create_task(
                        duplicate.get_response(), name="duplicate.get_response"
                    )
                    done_futs = await first_completed(fut, duplicate_task, cancel=True)
                    # cancel if not finished, suppress exc logging if finished
                    duplicate_task.cancel()
                    duplicate._fut.cancel()
                    for d in done_futs:
                        response = d.result()
                        if d is not fut:
                            # this means duplicate_task finished first
                            # in case fut is also finished with an exception, we'll mark it as retrieved
                            fut._Future__log_traceback = False
                            return response

        except ClientResponseError as e:
            # TODO think about getting rid of this
            raise DankMidsClientResponseError(e, self.request) from e
        except Exception as e:
            if not hasattr(e, "request"):
                e.request = request = self.request  # type: ignore [attr-defined]
                e.args = *e.args, request
            raise

        if not isinstance(response, RawResponse):
            return response

        # JIT json decoding
        response = response.decode(partial=True)

        if response.error is None:
            if self.raw and response.result:
                return {"result": response.result}
            return response.to_dict(self.method)

        if needs_full_request_spec(response) and self.controller._check_request_type():
            method: RPCEndpoint = f"{self.method}_raw" if self.raw else self.method  # type: ignore [assignment]
            return await self.controller(method, self.params)
        elif revert_logger.isEnabledFor(DEBUG) and type(response.exception) is ExecutionReverted:
            revert_logger_log_debug("%s for %s", response.exception, self)
        elif error_logger.isEnabledFor(DEBUG) and type(response.exception) is not ExecutionReverted:
            error_logger_log_debug("%s for %s", response.error, self)

        response = response.to_dict(self.method)
        response["error"] = dict(response["error"].items(), dankmids_added_context=self.request)
        return response

    async def get_response_unbatched(self) -> RPCResponse:  # type: ignore [override]
        task = create_task(self.make_request(), name="RPCRequest.get_response_unbatched")
        try:
            await wait_for(shield(task), timeout=TIMEOUT_SECONDS_BIG)
        except CancelledError:
            task.cancel()
            raise
        except TimeoutError:
            # looks like its stuck for some reason, let's try another one
            _log_debug(
                "%s got stuck in `get_response_unbatched`, we're creating a new one...",
                self,
            )

            duplicate = self.create_duplicate()

            # don't start counting for the timeout while we still have a queue of requests to send
            await rate_limit_inactive(self.controller.endpoint)

            duplicate_task = create_task(
                duplicate.get_response_unbatched(), name="duplicate.get_response_unbatched"
            )
            done: Set[Task] = await first_completed(task, duplicate_task, cancel=True)
            for fut in done:
                if fut is not task:
                    # this means the duplicate completed first
                    return fut.result()
            # cancel the duplicate if it wasn't the one that completed first
            duplicate._fut.cancel()
        response: RawResponse = await self._fut
        decoded = response.decode(partial=True)
        return (
            {"result": decoded.result}
            if self.raw and decoded.result
            else decoded.to_dict(self.method)
        )

    async def spoof_response(self, data: Union[RawResponse, bytes, Exception]) -> None:
        # sourcery skip: merge-duplicate-blocks
        """
        `Raw` type data comes from rpc calls executed in a jsonrpc batch
        `bytes` type data comes for individual eth_calls that were batched into multicalls and already decoded
        `Exception` type data comes from failed calls
        """

        # New handler
        if isinstance(data, RawResponse):
            self._fut.set_result(data)
        elif isinstance(data, BadResponse):
            if needs_full_request_spec(data.response) and self.controller._check_request_type():
                self._fut.set_result(await self.create_duplicate())
                return
            formatted = data.response.error.to_dict(context=self.request)  # type: ignore [union-attr]
            self._fut.set_result({"error": formatted})
            if error_logger.isEnabledFor(DEBUG):
                error_logger_log_debug("RPC Error for %s", self)
                error_logger_log_debug("response set: %s", formatted)
        elif isinstance(data, Exception):
            self.__set_exception(data)
        # From multicalls
        elif isinstance(data, bytes):
            self._fut.set_result({"result": data})
        else:
            dtype = type(data)
            exc = TypeError(f"{dtype.__name__} not supported for spoofing.", dtype, data)
            self.__set_exception(exc)

    async def make_request(self, num_previous_timeouts: int = 0) -> RawResponse:
        """
        Execute the request with no batching.

        Args:
            num_previous_timeouts (optional): the number of times this request has already been attempted and timed out. Default 0.
        """
        task = create_task(
            coro=self.controller.make_request(self.method, self.params, request_id=self.uid),
            name=f"RPCRequest.make_request attempt {num_previous_timeouts+1}",
        )
        try:
            response = await wait_for(shield(task), TIMEOUT_SECONDS_SMALL)
        except CancelledError:
            task.cancel()
            raise
        except TimeoutError:
            log_func = timeout_logger_warning if num_previous_timeouts > 1 else timeout_logger_debug
            log_func(
                "`make_request` timed out (%ss) %s times for %s, trying again...",
                TIMEOUT_SECONDS_SMALL,
                num_previous_timeouts + 1,
                self,
            )
            next_attempt_task = create_task(
                self.make_request(num_previous_timeouts + 1), name="next attempt task"
            )
            done: Set[Task] = await first_completed(task, next_attempt_task, cancel=True)
            first_done = done.pop()
            response = first_done.result()
            if first_done is not next_attempt_task:
                # next_attempt_task would have already set the fut result
                self._fut.set_result(response)
        else:
            self._fut.set_result(response)

        return response

    def create_duplicate(self) -> Union["RPCRequest", "eth_call"]:
        dupe_uid = f"{self.uid}_copy"
        if type(self) is eth_call:
            duplicate = eth_call(self.controller, self.params, dupe_uid)
        else:
            method: RPCEndpoint = f"{self.method}_raw" if self.raw else self.method  # type: ignore [assignment]
            duplicate = RPCRequest(self.controller, method, self.params, dupe_uid)

        selffut = self._fut
        dupefut = duplicate._fut

        def _self_done_callback(selffut):
            if dupefut.done():
                return

            dupefut.remove_done_callback(_dupe_done_callback)

            if selffut.cancelled():
                dupefut.cancel()
            elif (exc := selffut.exception()) is not None:
                dupefut.set_exception(exc)
            else:
                dupefut.set_result(selffut.result())

        def _dupe_done_callback(dupefut):
            if selffut.done():
                return

            selffut.remove_done_callback(_self_done_callback)

            if dupefut.cancelled():
                selffut.cancel()
            elif (exc := dupefut.exception()) is not None:
                selffut.set_exception(exc)
            else:
                selffut.set_result(dupefut.result())

        selffut.add_done_callback(_self_done_callback)
        dupefut.add_done_callback(_dupe_done_callback)

        return duplicate

    def __set_exception(self, data: Exception) -> None:
        if revert_logger.isEnabledFor(DEBUG) and type(data) in _REVERT_EXC_TYPES:
            self.__log_set_exception(revert_logger_log_debug, data)
        elif error_logger.isEnabledFor(DEBUG) and type(data) not in _REVERT_EXC_TYPES:
            self.__log_set_exception(error_logger_log_debug, data)
        self._fut.set_exception(data)

    def __log_set_exception(self, log_func: Callable[..., None], exc: Exception) -> None:
        exc_type = type(exc).__name__
        log_func("%s for %s", exc_type, self)
        log_func("exception set: %s", repr(exc))


_rpcrequest_init: Final = RPCRequest.__init__


@final
class eth_call(RPCRequest):
    revert_threads = PruningThreadPoolExecutor(4)

    __slots__ = "target", "calldata", "block"

    def __init__(
        self, controller: "DankMiddlewareController", params: Any, uid: Optional[str] = None
    ) -> None:
        """Adds a call to the DankMiddlewareContoller's `pending_eth_calls`."""

        call_dict, block = params

        self.target: ChecksumAddress = call_dict["to"]
        """The contract address for the call."""

        self.calldata: HexBytes = HexBytes(call_dict["data"])
        """The calldata for the call."""

        self.block: BlockId = block
        """The block height at which the contract will be called."""

        _rpcrequest_init(self, controller, "eth_call", params, uid)

    def __repr__(self) -> str:
        tx, block = self.params
        batch = self._batch
        batch_info = "" if batch is None else f" batch={batch}"
        if batch is None or type(batch) is not Multicall:
            if block.startswith("0x"):
                block = int(block, 16)
            block_info = f" block={block}"
        else:
            block_info = ""
        return f"<{self.__class__.__name__} uid={self.uid}{block_info} to={tx['to']} data={tx['data']}{batch_info}>"

    @property
    def multicall_compatible(self) -> bool:
        """True if this contract is multicall compatible, False if not."""
        return self.target not in self.controller.no_multicall

    async def spoof_response(self, data: Union[bytes, Exception, RawResponse]) -> None:  # type: ignore
        """Sets and returns a spoof rpc response for this BatchedCall instance using data provided by the worker."""

        # NOTE: If `type(data)` is `bytes`, it is a result from a multicall. If not, `data` comes from a jsonrpc batch.
        # If this if clause is True, it means the call reverted inside of a multicall but returned a result, without causing the multicall to revert.
        if is_revert_bytes(data):
            # TODO figure out how to include method selector in no_multicall key
            try:
                # NOTE: If call response from multicall indicates failure, make sync call to get either:
                # - successful response
                # - revert details from exception
                # If we get a successful response, most likely the target contract does not support multicall2.
                # TODO: Get rid of the sync executor and just use `make_request`
                controller = self.controller
                target = self.target
                failures = 0
                while failures < 5:
                    try:
                        data = await self.revert_threads.run(
                            controller.sync_w3.eth.call,
                            {"to": target, "data": self.calldata},
                            self.block,
                        )
                    except ReadTimeout:
                        failures += 1
                    else:
                        break
                # The single call was successful. We don't want to include this contract in more multicalls
                controller.no_multicall.add(target)
            except Exception as e:
                # NOTE: The call still returns a revert when it's not packed in a multicall
                data = e

        # The above revert catching logic fails to account for pre-decoding RawResponse objects.
        await RPCRequest.spoof_response(self, data)

    def _get_mc_data(self) -> MulticallChunk:
        return self.target, self.calldata


### Batch requests:

_Request = TypeVar("_Request", bound=_RequestBase)


class _Batch(_RequestBase[List[_Response]], Iterable[_Request]):
    calls: Final[WeakList[_Request]]
    _done: Final[_RequestEvent]

    _awaited: bool = False
    """A flag indicating whether the batch has been awaited."""

    __slots__ = "calls", "_batcher", "_lock", "_done", "_daemon", "__dict__"

    def __init__(self, controller: "DankMiddlewareController", calls: Iterable[_Request]) -> None:
        _request_base_init(self, controller)
        self.calls = WeakList(calls)
        self._batcher = controller.batcher
        self._lock = AlertingRLock(name=self.__class__.__name__)
        self._done = _RequestEvent(self)

    def __len__(self) -> int:
        return len(self.calls)

    def __await__(self) -> Generator[Any, None, _Response]:
        return self._task.__await__()

    @property
    def bisected(self) -> Generator[Tuple[_Request, ...], None, None]:
        # set `self.calls` output to var so its only computed once
        calls = tuple(self.calls)
        half = len(calls) // 2
        yield calls[:half]
        yield calls[half:]

    @property
    def is_full(self) -> bool:
        raise NotImplementedError(type(self).__name__)

    @cached_property
    def _task(self) -> "Task[None]":
        self._awaited = True
        return create_task(
            self.get_response(), name=f"{type(self).__name__} {self.uid} get_response"
        )

    def append(self, call: _Request, skip_check: bool = False) -> None:
        if self._awaited is True:
            raise RuntimeError(f"{self} was already awaited")
        with self._lock:
            self.calls.append(call)
            # self._len += 1
            if not skip_check:
                if self.is_full:
                    self.start()
                elif self.controller.queue_is_full:
                    self.controller.early_start()

    def extend(self, calls: Iterable[_Request], skip_check: bool = False) -> None:
        if self._awaited is True:
            raise RuntimeError(f"{self} was already awaited")
        with self._lock:
            self.calls.extend(calls)
            if not skip_check:
                if self.is_full:
                    self.start()
                elif self.controller.queue_is_full:
                    self.controller.early_start()

    def should_retry(self, e: Exception) -> bool:
        """Should the _Batch be retried based on `e`?"""
        if type(e) is OutOfGas:
            # TODO Remember which contracts/calls are gas guzzlers
            if len(self) > 1:
                gas_logger_debug("%s out of gas. cut in half, trying again...", self)
                return True
            return False

        str_e = f"{e}".lower()
        if any(err in str_e for err in constants.RETRY_ERRS):
            # TODO: use these exceptions to optimize for the user's node
            if len(self) > 1 and error_logger.isEnabledFor(DEBUG):
                error_logger_log_debug(
                    "Dank too loud. Bisecting %s and retrying.", type(self).__name__
                )
        elif "429" in str_e or isinstance(e, ChainstackRateLimitError):
            pass
        elif isinstance(e, BadResponse) and (
            needs_full_request_spec(e.response) or is_call_revert(e)
        ):
            pass
        elif all(err not in str_e for err in constants.TOO_MUCH_DATA_ERRS):
            _log_warning("unexpected %s in a %s: %s", type(e).__name__, type(self).__name__, e)
        return len(self) > 1

    def _record_failure(self, e: Exception, data: str) -> None:
        _debugging.failures.record(
            self.controller.chain_id,
            e,
            type(self).__name__,
            self.uid,
            len(self),
            data,
        )

    async def _debug_daemon(self) -> None:
        done = self._done.is_set
        # NOTE: _resonse works for RPCRequst and eth_call, _done works for _Batch classes
        i = 1
        while self and not done():
            await sleep(60)
            if self and not done():
                _log_debug(
                    "%s has not received data after %s minute%s", self, i, "" if i == 1 else "s"
                )
                i += 1


_batch_init: Final = _Batch.__init__


@final
class Multicall(_Batch[RPCResponse, eth_call]):
    method: Final = "eth_call"

    # NOTE: value comes from eth_utils.function_signature_to_4byte_selector("tryBlockAndAggregate(bool,(address,bytes)[])")
    fourbyte: Final = b"9\x95B\xe9"

    _awaited: bool
    """A flag indicating whether the Multicall has been awaited."""

    __slots__ = ("bid",)

    def __init__(
        self,
        controller: "DankMiddlewareController",
        calls: Iterable[eth_call] = [],
        bid: Optional[BatchId] = None,
    ) -> None:
        # sourcery skip: default-mutable-arg
        _batch_init(self, controller, calls)
        self.bid: Final = bid or self.controller.multicall_uid.next

    def __repr__(self) -> str:
        block = self.block
        if block.startswith("0x"):
            block = int(block, 16)
        batch = self._batch
        batch_info = "" if batch is None else f" batch={batch}"
        return f"<Multicall mid={self.bid} block={block} len={len(self)}{batch_info} awaited={self._awaited}>"

    def __iter__(self) -> Iterator[eth_call]:
        return iter(self.calls)

    def __bool__(self) -> bool:
        return bool(self.calls)

    def __del__(self) -> None:
        calls = list(self.calls)
        if not calls or self._done.is_set():
            return

        # The Multicall still has calls that haven't been garbage collected
        loop_is_closed = None
        logged = False
        for call in calls:
            if not call._fut.done():
                if loop_is_closed is None:
                    loop_is_closed = call._fut._loop.is_closed()

                if loop_is_closed:
                    return

                if logged is False:
                    error_logger.error("%s was garbage collected before finishing", self)
                    logged = True

                try:
                    call._fut.set_exception(
                        GarbageCollectionError(
                            f"{self} was garbage collected before finishing.",
                            f"{call} might hang indefinitely if I don't raise this exception, "
                            "which only exists to help debug an issue inside of dank mids. "
                            "Please show it to Bob.",
                        )
                    )
                except RuntimeError as e:
                    if str(e) != "no running event loop":
                        raise

    @cached_property
    def block(self) -> BlockId:
        try:
            return next(iter(self.calls)).block
        except StopIteration as e:
            raise EmptyBatch(
                f"{type(self).__name__} {self.uid} is empty and should not be processed."
            ) from e.__cause__

    @property
    def calldata(self) -> str:
        return (self.fourbyte + mcall_encode(map(eth_call._get_mc_data, self.calls))).hex()  # type: ignore [misc]

    @cached_property
    def mcall(self) -> MulticallContract:
        return self.controller._select_mcall_target_for_block(self.block)

    @property
    def target(self) -> ChecksumAddress:
        return self.mcall.address

    @property
    def params(self) -> JsonrpcParams:
        target = self.target
        params = [{"to": target, "data": f"0x{self.calldata}"}, self.block]
        if self.needs_override_code and not self.controller.state_override_not_supported:
            params.append({target: {"code": self.mcall.bytecode}})
        return params  # type: ignore [return-value]

    @property
    def request(self) -> Union[Request, PartialRequest]:
        return self.controller.request_type(method=self.method, params=self.params, id=self.uid)

    @property
    def is_full(self) -> bool:
        return len(self) >= self._batcher.step

    @property
    def needs_override_code(self) -> bool:
        return self.mcall.needs_override_code_for_block(self.block)

    def start(self, batch: Optional[Union["_Batch", "DankBatch"]] = None, cleanup=True) -> None:
        batch = batch or self
        if _logger_is_enabled_for(DEBUG):
            self._daemon = create_task(self._debug_daemon(), name="Multicall debug daemon")
        with self._lock:
            for call in self.calls:
                call._batch = self
            if cleanup:
                controller = self.controller
                with controller.pools_closed_lock:
                    controller._pending_eth_calls_pop(self.block, None)

    async def get_response(self) -> None:  # type: ignore [override]
        # create a strong ref to all calls we will execute so they cant get gced mid execution and mess up response ordering
        calls = tuple(self.calls)

        if not calls:
            # TODO: figure out how we get into this function without any calls
            self._done.set()
            return

        elif len(calls) == 1:
            await self._exec_single_call()
            self._done.set()
            return

        # elif l < 50: # TODO play with later
        #    return await JSONRPCBatch(self.controller, self.calls)

        controller = self.controller
        rid = controller.request_uid.next
        _demo_logger_info("request %s for multicall %s starting", rid, self.bid)
        try:
            await self.spoof_response(
                await controller.make_request(self.method, self.params, request_id=self.uid),
                calls,
            )
        except internal_err_types.__args__ as e:  # type: ignore [attr-defined]
            stre = str(e)
            if "invalid argument" in stre:
                raise
            elif "iteration" in stre:
                error_logger.error("ERROR: stop iteration in dank mids:")
                error_logger.exception(e)
            raise DankMidsInternalError(e) from e
        except Exception as e:
            if isinstance(e, ClientResponseError) and e.message == "Payload Too Large":
                _log_info("Payload too large. response headers: %s", e.headers)
                controller.reduce_multicall_size(len(self))
                self._record_failure(e, self.request.data.decode())
            elif _log_exception(e):
                self._record_failure(e, self.request.data.decode())

            if len(calls) == 1:
                await self._exec_single_call()
                self._done.set()
            elif self.should_retry(e):
                await self.bisect_and_retry(e)
            else:
                await self.spoof_response(e)

        _demo_logger_info("request %s for multicall %s complete", rid, self.bid)

    def should_retry(self, e: Exception) -> bool:
        """Should the Multicall be retried based on `e`?"""
        # NOTE: While it might look weird, f-string is faster than `str(e)`.
        if any(map(f"{e}".lower().__contains__, constants.RETRY_ERRS)):
            msg = "dank too loud, multicall %s got exc%s\ntrying again..."
            error_logger_debug(msg, self.bid, e)
            return True
        elif "No state available for block" in f"{e}":
            note = "You're not using an archive node, and you need one for the application you are attempting to run."
            e.args[0]["dankmids_note"] = note
            return False
        elif _Batch.should_retry(self, e):
            return True
        return len(self) > 1

    @set_done
    async def spoof_response(
        self, data: Union[RawResponse, Exception], calls: Optional[Sequence[eth_call]] = None
    ) -> None:
        # This happens if an Exception takes place during a non-batched Multicall request.
        if isinstance(data, Exception):
            if error_logger.isEnabledFor(DEBUG):
                exc_type = type(data).__name__
                error_logger_log_debug("%s had %s %s", self, exc_type, data)
                error_logger_log_debug("propagating the %s to all %s's calls", exc_type, self)

            # No need to gather this, `spoof_response` with an Exception input will complete synchronously
            for call in calls or self.calls:
                await call.spoof_response(data)

        # A `RawResponse` represents either a successful or a failed response, stored as pre-decoded bytes.
        # It was either received as a response to a single rpc call or as a part of a batch response.
        elif isinstance(data, RawResponse):
            response = data.decode(partial=True)
            if response.error is not None:
                exc = response.exception
                if error_logger.isEnabledFor(DEBUG) and type(exc) is not OutOfGas:
                    error_logger_log_debug(
                        "%s for %s",
                        response.error if type(exc) is BadResponse else repr(exc),
                        self,
                    )
                # NOTE: We raise the exception which will be caught, call will be broken up and retried
                raise exc
            _log_debug("%s received valid bytes from the rpc", self)

            # We write some ugly code to separate successes from reverts
            # For successful calls, we can set the result right away
            # For reverts, we asynchronously attempt to handle the revert

            # NOTE: we pass in the calls to create a strong reference so when we zip up the results everything gets to the right place
            if calls is None:
                calls = tuple(self.calls)

            to_gather = []
            for call, result in zip(calls, await self.decode(response)):
                if is_revert_bytes(result):
                    # We will asynchronously handle this revert
                    to_gather.append(eth_call.spoof_response(call, result))
                else:
                    # `spoof_response` with a successful call result will complete synchronously
                    await eth_call.spoof_response(call, result)

            await gatherish(to_gather, name="Multicall.spoof_response gatherish")

        else:
            raise NotImplementedError(f"type {type(data)} not supported.", data)

    async def decode(self, data: PartialResponse) -> List[bytes]:
        start = time()
        if ENVS.OPERATION_MODE.infura or len(self) < 100:
            # decode synchronously
            retval = mcall_decode(data)
        else:
            try:
                retval = await ENVS.MULTICALL_DECODER_PROCESSES.run(mcall_decode, data)  # type: ignore [attr-defined]
            except BrokenProcessPool:
                # TODO: Move this somewhere else
                logger.critical("Oh fuck, you broke the %s while decoding %s", ENVS.MULTICALL_DECODER_PROCESSES, data)  # type: ignore [attr-defined]
                ENVS.MULTICALL_DECODER_PROCESSES = AsyncProcessPoolExecutor(ENVS.MULTICALL_DECODER_PROCESSES._max_workers)  # type: ignore [attr-defined,assignment]
                retval = mcall_decode(data)

        stats.log_duration(f"multicall decoding for {len(self)} calls", start)
        # Raise any Exceptions that may have come out of the process pool.
        if isinstance(retval, Exception):
            raise retval.__class__(
                *retval.args,
                self.request,
                f"response: {data}",
            )
        return retval

    @set_done
    async def bisect_and_retry(self, e: Exception) -> None:
        """
        Split the :class:`~Multicall` into 2 chunks, then await both.

        Calls `self._done.set()` when finished.

        Args:
            e: The Exception that occured to cause the retry.
        """
        if error_logger.isEnabledFor(DEBUG) and type(e) is not OutOfGas:
            error_logger_log_debug(
                "%s had %s, bisecting and retrying...",
                self,
                e.response.error if type(e) is BadResponse else repr(e),
            )
        controller = self.controller
        # we need to create strong refs to the multicalls here so they dont disappear as soon as the JSONRPCBatch inits
        bisected = (batch0, batch1) = [
            Multicall(controller, chunk, f"{self.bid}_{i}") for i, chunk in enumerate(self.bisected)
        ]
        if controller.pending_rpc_calls:
            controller._pending_rpc_calls_append(batch0)
            if batch1:
                if controller.pending_rpc_calls:
                    controller._pending_rpc_calls_append(batch1)
                else:
                    await batch1
            await controller.pending_rpc_calls._done.wait()
        elif batch1:
            batch = JSONRPCBatch(controller, bisected, f"{self.uid}_bisected")
            batch.start(self, cleanup=False)
            await batch
        elif batch0:
            batch0.start(self, cleanup=False)
            await batch0

    async def _exec_single_call(self) -> None:
        await next(iter(self.calls)).make_request()

    async def _spoof_or_retry(self, response: RawResponse) -> None:
        try:
            await self.spoof_response(response)
        except Exception as e:
            if not self.should_retry(e):
                raise
            await self.bisect_and_retry(e)


@final
class JSONRPCBatch(_Batch[RPCResponse, Union[Multicall, eth_call, RPCRequest]]):
    """
    Represents a batch of JSON-RPC requests.

    This class manages a collection of JSON-RPC requests that will be sent
    as a single batch to an Ethereum node, improving efficiency by reducing
    the number of separate network calls.
    """

    __slots__ = ("jid",)

    def __init__(
        self,
        controller: "DankMiddlewareController",
        calls: Iterable[Union[Multicall, RPCRequest]] = [],
        jid: Optional[BatchId] = None,
    ) -> None:  # sourcery skip: default-mutable-arg
        """
        Initialize a new JSONRPCBatch.

        Args:
            controller: The :class:`~DankMiddlewareController` instance that manages this batch.
            calls: A list of :class:`~RPCRequest` or :class:`~Multicall` objects to be included in the batch.
            jid: A unique identifier for this batch. If none is provided, one will be created.
        """
        _batch_init(self, controller, calls)
        self.jid = jid or self.controller.jsonrpc_batch_uid.next

    def __repr__(self) -> str:
        batch = self._batch
        batch_info = "" if batch is None else f" batch={batch}"
        return f"<JSONRPCBatch jid={self.jid} len={len(self)}{batch_info} awaited={self._awaited}>"

    def __iter__(self) -> Iterator[Union[Multicall, eth_call, RPCRequest]]:
        return filter(None, self.calls)

    def __bool__(self) -> bool:
        for _ in self:
            return True
        return False

    def __del__(self) -> None:
        if any(self) and not self._done.is_set():
            for cls, calls in groupby(self.calls, type):
                if cls is Multicall:
                    calls = concat(filter(None, calls))
                if any(filterfalse(Future.done, (call._fut for call in calls))):
                    error_logger.error("%s was garbage collected before finishing", self)
                    return

    @property
    def data(self) -> bytes:
        try:
            return b"[" + b",".join(call.request.data for call in self) + b"]"
        except TypeError as e0:
            # If we can't encode one of the calls, lets figure out which one and pass some useful info downstream
            for call in self:
                try:
                    call.request.data
                except TypeError as e1:
                    raise TypeError(e1, call.request) from e0.__cause__
            raise

    @property
    def is_multicalls_only(self) -> bool:
        with self._lock:
            return all(isinstance(call, Multicall) for call in self)

    @property
    def is_single_multicall(self) -> bool:
        with self._lock:
            return len(self) == 1 and self.is_multicalls_only

    @property
    def method_counts(self) -> Dict[RPCEndpoint, int]:
        """
        Count the occurrences of each method in the batch.

        Returns:
            A dictionary where keys are method names and values are their counts.
        """
        counts: DefaultDict[RPCEndpoint, int] = defaultdict(int)
        with self._lock:
            for call in self:
                if isinstance(call, Multicall):
                    counts["eth_call[multicall]"] += len(call)  # type: ignore
                else:
                    counts[call.method] += 1
            return dict(counts)

    @property
    def total_calls(self) -> int:
        """
        Get the total number of calls across all requests in the batch.

        This may differ from __len__ if the batch includes multicalls.

        Returns:
            The total number of individual calls in the batch.
        """
        with self._lock:
            return sum(map(len, self))

    @property
    def is_full(self) -> bool:
        with self._lock:
            return (
                self.total_calls >= self._batcher.step
                or len(self) >= self.controller.max_jsonrpc_batch_size
            )

    def start(self, batch: Optional["DankBatch"] = None, cleanup=True) -> None:
        # sourcery skip: hoist-loop-from-if
        batch = batch or self
        if _logger_is_enabled_for(DEBUG):
            self._daemon = create_task(self._debug_daemon(), name="JSONRPCBatch debug daemon")
        with self._lock:
            for typ, calls in groupby(self.calls, type):
                if typ is Multicall:
                    for call in calls:
                        call.start(batch, cleanup=False)
                else:
                    for call in calls:
                        call._batch = self
            if cleanup:
                self.controller._start_new_batch()

    async def get_response(self) -> None:  # type: ignore [override]
        if not self.calls:
            # TODO: figure out why this can happen and prevent it upstream
            self._done.set()
            return

        if self.is_single_multicall:
            multicall = next(iter(self.calls))
            await multicall
            self._done.set()
            return

        rid = self.controller.request_uid.next
        if ENVS.DEMO_MODE:  # type: ignore [attr-defined]
            # When demo mode is disabled, we can save some CPU time by skipping this sum
            _demo_logger_info(
                "request %s for jsonrpc batch %s (%s calls) starting",
                rid,
                self.jid,
                sum(map(len, self)),
            )
        try:
            while True:
                try:
                    # NOTE: We do this inline so we never have to allocate the response to memory
                    await self.spoof_response(*await self.post())
                    break
                except ChainstackRateLimitError as e:
                    # Chainstack doesn't use 429 for rate limiting, it sends a successful response back to the rpc with an error
                    # message so our usual rate-limiting handlers don't work and we need to handle that case with bespoke logic.
                    await sleep(e.try_again_in)
                except RateLimitError:
                    # Quiknode doesn't use 429 for rate limiting, it sends a successful response back to the rpc with an error
                    # message so our usual rate-limiting handlers don't work and we need to handle that case with bespoke logic.
                    logger.warning("rate limited by quiknode")
                    await sleep(5)

        # I want to see these asap when working on the lib.
        except internal_err_types.__args__ as e:  # type: ignore [attr-defined]
            raise e if "invalid argument" in str(e) else DankMidsInternalError(e) from e
        except EmptyBatch:
            self._done.set()
        except ExceedsMaxBatchSize as e:
            _log_warning("exceeded max batch size for your node")
            self.controller.set_batch_size_limit(e.limit)
            await self.bisect_and_retry(e)
        except PayloadTooLarge as e:
            # TODO: track too large payloads and do some better optimizations for batch sizing
            self.adjust_batch_size()
            await self.bisect_and_retry(e)
        except Exception as e:
            if _log_exception(e):
                self._record_failure(e, self.data)

            stats.log_errd_batch(self)

            if self.should_retry(e):
                # should_retry will always return True if there is more than 1 call in this batch
                await self.bisect_and_retry(e)

            else:
                # NOTE: This means an exception occurred during the post request
                # AND that the json batch is made of just one rpc request that is not a multicall.
                error_logger_debug(
                    "%s had exception %s, aborting and sending Exception to waiters", self, e
                )
                await igather(call.spoof_response(e) for call in self.calls)

        _demo_logger_info("request %s for jsonrpc batch %s complete", rid, self.jid)  # type: ignore

    @eth_retry.auto_retry(min_sleep_time=0, max_sleep_time=1, suppress_logs=2)
    async def post(self) -> Tuple[List[RawResponse], List[Union[Multicall, RPCRequest]]]:
        """
        Send the batch of requests to the Ethereum node and process the responses.

        This method sends the batch of requests to the Ethereum node and
        processes the responses.

        Returns:
            A tuple containing the raw responses and the list of calls in the batch.

        Raises:
            BadResponse: If a successful 'error' response was received from the rpc.
            ClientResponseError: If there was an error with the HTTP request.
            Exception: For other unexpected errors.
        """
        try:
            # we need strong refs so the results all get to the right place
            calls = tuple(self)
            if not calls:
                # TODO: figure out why this can happen and prevent it elsewhere
                return [], []

            # for the multicalls too
            mcall_calls_strong_refs = tuple(tuple(call.calls) for call in calls if type(call) is Multicall)  # type: ignore [union-attr]
            response: JSONRPCBatchResponse = await _session.post(
                self.controller.endpoint, data=self.data, loads=_codec.decode_jsonrpc_batch
            )
        except ClientResponseError as e:
            if e.message == "Payload Too Large":
                _log_warning("Payload too large: %s", self.method_counts)
                self.adjust_batch_size()
            elif "broken pipe" in str(e).lower():
                _log_warning("This is what broke the pipe: %s", self.method_counts)
            error_logger_debug("caught %s for %s, reraising", e, self)
            if ENVS.DEBUG:  # type: ignore [attr-defined]
                self._record_failure(e, self.data.decode())
            raise
        except Exception as e:
            if "broken pipe" in str(e).lower():
                _log_warning("This is what broke the pipe: %s", self.method_counts)
            if ENVS.DEBUG:  # type: ignore [attr-defined]
                self._record_failure(e, self.data.decode())
            raise

        # NOTE: A successful response will be a list of `RawResponse` objects.
        #       A single `PartialResponse` implies an error.
        if isinstance(response, list):
            return response, calls

        # Oops, we failed.
        errmsg = response.error.message
        if errmsg.lower() in ("invalid request", "parse error"):  # type: ignore [union-attr]
            # NOT SURE IF THIS ACTUALLY RUNS, CAN WE RECEIVE THIS TYPE RESPONSE FOR A JSON BATCH?
            self.controller._check_request_type()
        elif errmsg == "batch limit exceeded":  # type: ignore [union-attr]
            self.adjust_batch_size()

        if ENVS.DEBUG:
            self._record_failure(response.exception)

        raise response.exception

    def should_retry(self, e: Exception) -> bool:
        """
        Determine whether the JSONRPCBatch should be retried based on the exception.

        Args:
            e: The exception that occurred during the batch request.

        Returns:
            True if the batch should be retried, False otherwise.
        """
        # While it might look weird, f-string is faster than `str(e)`.
        if "No state available for block" in f"{e}":
            error_logger_debug(
                "No state available for one of the blocks queried. Bisecting batch and retrying."
            )
            return True
        elif self.is_single_multicall:
            # TODO: delete this eventually if it doesn't trigger
            raise RuntimeError(
                "pretty sure we can't get here anymore now that I check this in the beginning of JSONRPCBatch.get_response"
            )
        return _Batch.should_retry(self, e)

    @set_done
    async def spoof_response(
        self, response: List[RawResponse], calls: Tuple[RPCRequest, ...]
    ) -> None:
        """
        Process the responses from the Ethereum node and set the results for each call.

        This method processes the responses from the Ethereum node and sets the results
        for each call in the batch. It also handles any exceptions that occurred during
        the processing.

        Args:
            response: A list of RawResponse objects containing the responses from the node.
            calls: A list of RPCRequest objects that were included in the batch.
        """
        # Reaching this point means we made a batch call and we got results. That doesn't mean they're good, but we got 'em.

        controller = self.controller

        if controller._sort_calls:
            # NOTE: these providers don't always return batch results in the correct ordering
            # NOTE: is it maybe because they
            try:
                calls = sorted(calls, key=lambda call: call.uid)  # type: ignore [assignment]
            except TypeError:
                # '<' not supported between instances of 'int' and 'str'
                # This happens when a multicall or jsonrpc batch was split in half.
                # Those bisected batches have a string uid to associate them with
                # the original.
                calls = sorted(calls, key=lambda call: str(call.uid))

        if controller._sort_response:
            try:
                response.sort(key=lambda raw: raw.decode().id)
            except TypeError:
                # '<' not supported between instances of 'int' and 'str'
                # This happens when a multicall or jsonrpc batch was split in half.
                # Those bisected batches have a string uid to associate them with
                # the original.
                response.sort(key=lambda raw: str(raw.decode().id))
        else:
            for call, raw in zip(calls, response):
                # TODO: make sure this doesn't ever raise and then delete it
                if call.uid != raw.decode().id:
                    # Not sure why it works this way
                    raise BatchResponseSortError(controller, calls, response)

        responses = iter(response)
        mcall_coros = []
        for request_type, requests in groupby(calls, type):
            if request_type is Multicall:
                mcall_coros.extend(map(Multicall._spoof_or_retry, requests, responses))
            else:
                # These do not need to be delegated to asks since they
                # will always complete synchronously when called here
                for coro in map(request_type.spoof_response, requests, responses):
                    await coro

        if mcall_coros:
            await gatherish(mcall_coros, name="JSONRPCBatch.spoof_response gatherish")

    @set_done
    async def bisect_and_retry(self, e: Exception) -> None:
        """
        Split the batch into two halves and retry each half separately.

        This method splits the batch of calls into two halves and retries each half
        separately. If one half is just a single multicall, it will be handled alone,
        not placed into a batch.

        Sets :attr:`~JSONRPCBatch._done` on the batch when finished.

        Args:
            e: The exception that occurred during the batch request.
        """
        if error_logger.isEnabledFor(DEBUG):
            exc_str = e.response.error if type(e) is BadResponse else repr(e)
            error_logger_log_debug("%s had %s", self, exc_str)
            error_logger_log_debug("retrying...")

        controller = self.controller
        batches = [
            (
                Multicall(controller, first_call, f"json{self.jid}_{i}")  # type: ignore [misc]
                if len(chunk) == 1 and isinstance(first_call := next(iter(chunk)), Multicall)
                else JSONRPCBatch(controller, chunk, f"{self.jid}_{i}")
            )
            for i, chunk in enumerate(self.bisected)
            if chunk
        ]
        for batch in batches:
            batch.start(self, cleanup=False)
        for batch, result in zip(batches, await igather(batches, return_exceptions=True)):
            if isinstance(result, Exception):
                if not isinstance(result, DankMidsInternalError):
                    log_internal_error(logger, batch, result)
                raise result

    def adjust_batch_size(self) -> None:
        """
        Adjust the batch size on the controller based on the type of calls in the batch.

        This method adjusts the batch size based on whether the batch contains
        multicalls or regular JSON-RPC calls. It logs appropriate messages and
        updates the batch size accordingly.
        """
        if self.is_multicalls_only:
            num_calls = self.total_calls
            batch_size.log_check("multicall", "calls", num_calls)
            self.controller.reduce_multicall_size(num_calls)
        else:
            batch_size.log_check("json", "requests", len(self))
            self.controller.reduce_batch_size(len(self))
            _log_devhint(
                "We still need some better logic for catching these errors and using them to better optimize the batching process"
            )


# NOTE: These errors are expected during normal use and are not indicative of any problem(s). No need to log them.
_DONT_NEED_TO_SEE_ERRS = [
    "non_empty_data",
    "exceeding --rpc.returndata.limit",
    "'code': 429",
]


def _log_exception(e: Exception) -> bool:
    """
    Log exceptions that occur during a multicall or batch.

    This function logs exceptions that are unexpected and considered errors,
    allowing for better debugging and monitoring.

    Args:
        e: The exception to log.

    Returns:
        True if the exception should be logged, False otherwise.
    """
    # NOTE: These errors are expected during normal use and are not indicative of any problem(s). No need to log them.
    if type(e) is OutOfGas:
        return ENVS.DEBUG

    # TODO: Better filter what we choose to log here

    dont_need_to_see_errs = [
        *_DONT_NEED_TO_SEE_ERRS,
        # We catch and correct these
        "invalid request",
        # We pass these down to the call they originated from
        *INDIVIDUAL_CALL_REVERT_STRINGS,
    ]

    stre = str(e).lower()
    if all(
        err not in stre
        for err in chain(dont_need_to_see_errs, constants.RETRY_ERRS, constants.TOO_MUCH_DATA_ERRS)
    ):
        _log_warning(
            "The following exception is being logged for informational purposes and does not indicate failure:"
        )
        _log_warning(e, exc_info=True)
    return ENVS.DEBUG  # type: ignore [attr-defined,return-value]


_log_debug: Final = logger.debug
_log_info: Final = logger.info
_log_warning: Final = logger.warning
_logger_is_enabled_for: Final = logger.isEnabledFor
_log_devhint: Final = stats.logger.devhint
_demo_logger_info: Final = demo_logger.info



from functools import wraps
from typing import Callable, Dict, TypeVar

from typing_extensions import ParamSpec


__P = ParamSpec("__P")
__T = TypeVar("__T")


def lru_cache_lite(func: Callable[__P, __T]) -> Callable[__P, __T]:
    """A faster implementation of functools.lru_cache with no maxsize.

    This decorator caches the results of the decorated function without imposing a
    maximum cache size. It uses a unique internal cache-miss marker to determine if a
    result is cached.

    **Example:**

        .. code-block:: python

            @lru_cache_lite
            def add(x: int, y: int) -> int:
                print("Computing add...")
                return x + y

            # First call with same arguments results in function execution.
            print(add(2, 3))  # prints "Computing add..." then 5

            # Subsequent calls with the same arguments return the cached result.
            print(add(2, 3))  # prints 5 without "Computing add..."

    **See Also:**
        - `functools.lru_cache <https://docs.python.org/3/library/functools.html#functools.lru_cache>`_
        - :func:`~dank_mids.lru_cache_lite_nonull` for a variant using ``None`` as the cache-miss marker.
    """
    cache: Dict[__P.args, __T] = {}  # type: ignore [valid-type]
    cache_miss = object()

    @wraps(func)
    def lru_cache_lite_wrap(*args: __P.args) -> __T:  # type: ignore [valid-type]
        retval = cache.get(args, cache_miss)
        if retval is cache_miss:
            retval = func(*args)  # type: ignore [call-arg]
            cache[args] = retval
        return retval  # type: ignore [return-value]

    lru_cache_lite_wrap.cache = cache  # type: ignore [attr-defined]
    return lru_cache_lite_wrap  # type: ignore [return-value]


def lru_cache_lite_nonull(func: Callable[__P, __T]) -> Callable[__P, __T]:
    """A faster implementation of functools.lru_cache with no maxsize that uses None as the cache miss indicator.

    This decorator behaves similarly to :func:`~dank_mids.lru_cache_lite` but differs by using
    ``None`` as the default value returned by the cache when a key is missing.
    Consequently, if the decorated function legitimately returns ``None`` for a given input,
    the cache will interpret the cached ``None`` as a miss and will re-evaluate the function.

    **Example:**

        .. code-block:: python

            @lru_cache_lite_nonull
            def might_return_none(x: int) -> int:
                print("Executing might_return_none...")
                return x if x > 0 else None

            # First call with x = -1 triggers a cache miss and function execution.
            result = might_return_none(-1)
            print(result)  # prints "Executing might_return_none..." then None

            # Second call with x = -1 also triggers a cache miss because None is used as the cache-miss marker.
            result = might_return_none(-1)
            print(result)  # prints "Executing might_return_none..." then None

    **Warnings:**
        Use :func:`~dank_mids.lru_cache_lite_nonull` only when it is acceptable that a cached return value
        of ``None`` results in re-execution of the decorated function. If the function may validly return ``None``,
        consider using :func:`~dank_mids.lru_cache_lite` instead.

    **See Also:**
        - :func:`~dank_mids.lru_cache_lite` which uses a unique cache-miss marker.
        - `functools.lru_cache <https://docs.python.org/3/library/functools.html#functools.lru_cache>`_
    """
    cache: Dict[__P.args, __T] = {}  # type: ignore [valid-type]

    @wraps(func)
    def lru_cache_lite_wrap(*args: __P.args) -> __T:  # type: ignore [valid-type]
        retval = cache.get(args)
        if retval is None:
            retval = func(*args)  # type: ignore [call-arg]
            cache[args] = retval
        return retval

    lru_cache_lite_wrap.cache = cache  # type: ignore [attr-defined]
    return lru_cache_lite_wrap  # type: ignore [return-value]


__all__ = ["lru_cache_lite", "lru_cache_lite_nonull"]



import logging
import re
from typing import TYPE_CHECKING, Final, List, Sequence, Union, final

from aiohttp.client_exceptions import ClientResponseError

if TYPE_CHECKING:
    from dank_mids._requests import RPCRequest
    from dank_mids.controller import DankMiddlewareController
    from dank_mids.helpers._codec import RawResponse
    from dank_mids.types import PartialRequest, PartialResponse


logger: Final = logging.getLogger("dank_mids.exceptions")


class BadResponse(ValueError):
    """Exception raised when an unexpected or invalid response is received."""

    def __init__(self, response: "PartialResponse") -> None:
        self.response = response
        BaseException.__init__(self, response.error.to_dict())  # type: ignore [union-attr]


@final
class EmptyBatch(ValueError):
    """Exception raised when attempting to process an empty batch."""


@final
class ExecutionReverted(BadResponse):
    """Exception raised when the EVM reverts."""

    def __init__(self, response: "PartialResponse") -> None:
        self.response = response
        message: str = response.error.message  # type: ignore [union-attr]
        if message.count(":") == 1:
            message = message.split(":")[1]
        BaseException.__init__(self, message)


@final
class OutOfGas(BadResponse):
    """Exception raised when an eth_call request runs out of gas."""


@final
class PayloadTooLarge(BadResponse):
    """Exception raised when the payload for a request is too large."""


@final
class ExceedsMaxBatchSize(BadResponse):
    """Exception raised when a batch operation exceeds the maximum allowed size for the RPC."""

    @property
    def limit(self) -> int:
        """Get the maximum allowed batch size from the error message."""
        return int(re.search(r"batch limit (\d+) exceeded", self.response.error.message)[1])  # type: ignore [index, union-attr]


@final
class DankMidsClientResponseError(ClientResponseError):  # type: ignore [misc]
    """A wrapper around the standard aiohttp ClientResponseError that attaches the request that generated the error."""

    def __init__(
        self,
        exc: ClientResponseError,
        request: "PartialRequest",
    ) -> None:
        self.request: Final = request
        """
        The PartialRequest object associated with this error.
        It contains information about the request that led to this error.
        """

        self.args = (*self.args, request)  # type: ignore [has-type]

        self._exception: Final = exc
        """
        The original ClientResponseError that this class is wrapping.
        It contains the underlying error information from aiohttp.
        """

        super().__init__(
            exc.request_info,
            exc.history,
            status=exc.status,
            message=exc.message,
            headers=exc.headers,
        )


_internal_err_types = Union[
    AttributeError, TypeError, UnboundLocalError, NotImplementedError, RuntimeError, SyntaxError
]
internal_err_types: Final = _internal_err_types


class DankMidsInternalError(Exception):
    """Exception raised for unhandled internal errors within Dank Mids."""

    def __init__(self, exc: Union[ValueError, _internal_err_types]) -> None:
        logger.warning(f"unhandled exception inside dank mids internals: {exc}", exc_info=True)

        self._original_exception: Final = exc
        """
        The original exception that was caught and wrapped by this DankMidsInternalError.
        This attribute allows access to the underlying error for debugging purposes.
        """

        super().__init__(repr(exc))


@final
class BatchResponseSortError(Exception):
    """
    Exception raised when there is an issue with sorting batch responses.
    Indicates special handling is needed for batch calls/responses in your RPC.
    """

    def __init__(
        self,
        controller: "DankMiddlewareController",
        calls: Sequence["RPCRequest"],
        response: List["RawResponse"],
    ) -> None:
        self.calls: Final = calls
        """
        A list of RPCRequest objects representing the calls that were made in the batch.
        This can be used to analyze which calls were included in the problematic batch.
        """

        self.results: Final = [raw.decode() for raw in response]
        """
        A list of Response objects representing the decoded results from the RPC call.
        This can be used to analyze the responses that couldn't be properly sorted.
        """

        super().__init__(
            f"This will not mess up your run but will make things needlessly slow. Please show this to Bob.\n"
            f"endpoint={controller.endpoint}\n"
            f"client_version={controller.client_version}\n"
            f"calls={[call.uid for call in calls]}\n"
            f"response={self.results}"
        )


class RateLimitError(BadResponse): ...


@final
class ChainstackRateLimitError(RateLimitError):
    """
    Exception raised when Chainstack hits a rate limit without using a 429 status code.
    Chainstack sends a 200 status code with an error message for rate limiting.
    Requires bespoke logic for handling this case.
    """

    @property
    def try_again_in(self) -> float:
        """
        Parses the response to determine the time to wait before retrying the request.

        Returns:
            The number of seconds to wait before retrying.
        """
        return self.response.error.data.try_again_in  # type: ignore [union-attr]


@final
class QuiknodeRateLimitError(RateLimitError): ...



from typing import Any

from web3.datastructures import AttributeDict


def make_hashable(obj: Any) -> Any:
    """
    Converts an object into a hashable type if possible.

    This function is used internally to ensure that objects can be used as
    dictionary keys or in sets.

    Args:
        obj: The object to make hashable.
    """
    if isinstance(obj, (list, tuple)):
        return tuple(make_hashable(o) for o in obj)
    elif isinstance(obj, dict):
        return AttributeDict({key: make_hashable(obj[key]) for key in obj})
    return obj



"""
This module provides debugging tools and utilities for dank_mids operations.

While currently private, this module is intended to be made public in the future
when the API stabilizes. It offers various debugging features to help developers
understand and troubleshoot dank_mids behavior.

Submodules:
- failures: Contains tools for logging and analyzing failed requests.

Note: The API in this module is subject to change. Use with caution in production environments.
"""

from dank_mids._debugging import failures

__all__ = ["failures"]



from logging import Logger
from typing import TYPE_CHECKING, Any, Final

from web3.exceptions import ContractLogicError

from dank_mids._exceptions import BadResponse
from dank_mids._logging import DEBUG, getLogger
from dank_mids.constants import REVERT_SELECTORS
from dank_mids.types import PartialResponse

if TYPE_CHECKING:
    from dank_mids._requests import _Batch


error_logger: Final = getLogger("dank_mids.errors")
error_logger_debug: Final = error_logger.debug
error_logger_log: Final = error_logger._log


def error_logger_log_debug(msg: str, *args: Any) -> None:
    error_logger_log(DEBUG, msg, args)


revert_logger: Final = error_logger.getChild("reverts")
revert_logger_log: Final = revert_logger._log


def revert_logger_log_debug(msg: str, *args: Any) -> None:
    revert_logger_log(DEBUG, msg, args)


timeout_logger: Final = error_logger.getChild("timeouts")
timeout_logger_debug: Final = timeout_logger.debug
timeout_logger_warning: Final = timeout_logger.warning


gas_logger: Final = error_logger.getChild("gas")
gas_logger_debug: Final = gas_logger.debug

INDIVIDUAL_CALL_REVERT_STRINGS: Final = {
    "invalid opcode",
    "missing trie node",
    "resource not found",
    "invalid ether transfer",
    "error processing call revert",
}


def log_internal_error(logger: Logger, batch: "_Batch", exc: Exception) -> None:  # type: ignore [type-arg]
    try:
        batch_objs = list(batch)
    except TypeError:
        # 'coroutine' object is not iterable
        if isinstance(exc, ContractLogicError):
            # we don't need to log reverts for single calls
            return
        batch_objs = [batch]
    logger.error(
        "That's not good, there was an exception in a %s (len=%s). These are supposed to be handled.\n"
        "Exc: %s\n"
        "%s contents: %s\n\n",
        type(batch).__name__,
        len(batch_objs),
        exc,
        type(batch).__name__,
        batch_objs,
        exc_info=True,
    )


def needs_full_request_spec(response: PartialResponse) -> bool:
    """
    Determine if a response indicates that the node requires the full request specification.

    By default we leave off some fields that are not always required.
    Some nodes do not like this, and they let us know via these errors.

    Args:
        response: The error response to check.

    Returns:
        True if the full request specification is needed, False otherwise.
    """
    return response.error is not None and response.error.message.lower() in (
        "invalid request",
        "parse error",
    )


def is_call_revert(e: BadResponse) -> bool:
    """
    Determine if a BadResponse was caused by a revert in one of the individual calls within a multicall.

    Args:
        e: The error response to check.

    Returns:
        True if the error was caused by an individual call revert, False otherwise.
    """
    stre = f"{e}".lower()
    return any(string in stre for string in INDIVIDUAL_CALL_REVERT_STRINGS)


def is_revert_bytes(data: Any) -> bool:
    """
    Check if a call reverted inside of a multicall but returned a result, without causing the multicall to revert.

    Args:
        data: The response to check.

    Returns:
        True if the call reverted, False if it was successful.
    """
    return isinstance(data, bytes) and any(
        data.startswith(selector) for selector in REVERT_SELECTORS
    )


def log_request_type_switch() -> None:
    error_logger_debug(
        "your node says the partial request was invalid but its okay, we can use the full jsonrpc spec instead"
    )



"""
stats.py

This module provides functionality for logging and collecting statistics related to the Dank Mids library.
It includes custom logging levels, a specialized logger, and classes for collecting and writing statistics.

Key components:
- Custom logging levels (STATS, DEVHINT)
- _StatsLogger: A specialized logger for Dank Mids statistics
- _Collector: Handles collection and computation of stats-related data
- _Writer: Converts collected stats into human-readable format
- _SentryExporter: Pushes metrics to Sentry

This module is crucial for debugging, performance monitoring, and optimization of the Dank Mids library.
"""

# TODO: Robust and Refactor

import logging
from asyncio import create_task
from collections import defaultdict, deque
from concurrent.futures import ProcessPoolExecutor
from copy import deepcopy
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Deque,
    Final,
    Iterable,
    Set,
    Tuple,
    Type,
    TypeVar,
    final,
)

import msgspec
from a_sync.asyncio import sleep0
from typed_envs.registry import _ENVIRONMENT_VARIABLES_SET_BY_USER
from web3.types import RPCEndpoint

from dank_mids import ENVIRONMENT_VARIABLES as ENVS
from dank_mids.stats import _nocompile

if TYPE_CHECKING:
    from dank_mids._requests import JSONRPCBatch
    from dank_mids.types import Request

_LogLevel = int

T = TypeVar("T")

# New logging levels:
# DEBUG=10, INFO=20,
STATS: Final = 13
"""Custom logging level for statistics, between DEBUG and INFO."""

DEVHINT: Final = 15
"""Custom logging level for developer hints, between STATS and INFO."""

COLLECT_STATS: bool = False  # TODO: enable this
"""Flag to enable or disable stats collection."""

_ENVS: Final = _ENVIRONMENT_VARIABLES_SET_BY_USER

yield_to_loop: Final = sleep0


# if you're both collecting data and logging something, put the function here:
# <func>


def log_errd_batch(batch: "JSONRPCBatch") -> None:
    """
    Log information about a failed JSON-RPC batch.

    Args:
        batch: The failed batch to log.
    """
    collector.errd_batches.append(batch)
    logger.devhint(
        f"jsonrpc batch failed\njson batch id: {batch.jid} | len: {len(batch)} | total calls: {batch.total_calls}\n"
        + f"methods called: {batch.method_counts}"
    )


def log_duration(work_descriptor: str, start: float, *, level: _LogLevel = STATS) -> None:
    """
    Log the duration of a specific operation.

    Args:
        work_descriptor: Description of the work being timed.
        start: Start time of the operation.
        level: Logging level to use. Defaults to STATS.
    """
    # sourcery skip: hoist-if-from-if
    enabled = logger.isEnabledFor(level)
    if COLLECT_STATS or enabled:
        duration = time() - start
        if COLLECT_STATS:
            collector.durations[work_descriptor].append(duration)
        if enabled:
            logger._log_nocheck(level, f"{work_descriptor} took {duration}")


@final
class _StatsLogger(logging.Logger):
    """
    A custom logger class for collecting and logging statistics about RPC method calls and responses.

    This logger extends the standard Python logging.Logger with methods for tracking validation errors,
    response types, and other statistics related to RPC interactions.
    """

    @property
    def enabled(self) -> bool:
        """Returns True if the logger is enabled for the STATS logging level (which is set to 13) or lower.

        This property checks if the current logging level is less than or equal to STATS and ensures the stats daemon
        is running. The STATS level is defined to be between DEBUG and INFO.

        Example:
            >>> _logger = _StatsLogger(__name__)
            >>> _logger.setLevel(STATS)
            >>> _logger.enabled
            True

        See Also:
            :meth:`isEnabledFor`
        """
        self._ensure_daemon()
        return self.isEnabledFor(STATS)

    # New logging levels

    def stats(self, msg: str, *args: Any, **kwargs: Any) -> None:
        """
        Log a message with the STATS logging level.

        Args:
            msg: The message to log.
            *args: Additional positional arguments for the logger.
            **kwargs: Additional keyword arguments for the logger.

        Example:
            >>> _logger.stats("Operation took %s seconds", 3.14)
        """
        if self.enabled:
            self._log_nocheck(STATS, msg, args, **kwargs)

    def devhint(self, msg: str, *args: Any, **kwargs: Any) -> None:
        """
        Log a message with the DEVHINT logging level.

        Args:
            msg: The message to log.
            *args: Additional positional arguments for the logger.
            **kwargs: Additional keyword arguments for the logger.

        Example:
            >>> _logger.devhint("This is a developer hint for debugging purposes")
        """
        self._log(DEVHINT, msg, args, **kwargs)

    # Functions to print stats to your logs.

    def log_brownie_stats(self, *, level: _LogLevel = STATS) -> None:
        """
        Log statistics related to Brownie operations.

        Args:
            level: The logging level to use. Defaults to STATS.

        Example:
            >>> _logger.log_brownie_stats()
        """
        self._log_fn_result(level, _Writer.brownie)

    def log_event_loop_stats(self, *, level: _LogLevel = STATS) -> None:
        """
        Log statistics about the event loop performance.

        Args:
            level: The logging level to use. Defaults to STATS.

        Example:
            >>> _logger.log_event_loop_stats()
        """
        self._log_fn_result(level, _Writer.event_loop)

    def log_subprocess_stats(self, *, level: _LogLevel = STATS) -> None:
        """
        Log statistics about subprocess pools and queues.

        Args:
            level: The logging level to use. Defaults to STATS.

        Example:
            >>> _logger.log_subprocess_stats()
        """
        for pool in {ENVS.BROWNIE_ENCODER_PROCESSES, ENVS.BROWNIE_DECODER_PROCESSES, ENVS.MULTICALL_DECODER_PROCESSES}:  # type: ignore [attr-defined]
            self._log_fn_result(level, _Writer.queue, pool)  # type: ignore [arg-type]

    # Internal helpers

    def _log(self, level: _LogLevel, msg: str, args: Tuple[Any, ...] = (), **kwargs: Any) -> None:  # type: ignore [override]
        """
        Wrapper around the standard logging method to simplify custom log level checks.

        Args:
            level: The logging level.
            *args: Additional positional arguments for the logger.
            **kwargs: Additional keyword arguments for the logger.
        """
        if self.isEnabledFor(level):
            return self._log_nocheck(level, msg, *args, **kwargs)

    def _log_nocheck(self, level: _LogLevel, msg: str, *args: Any, **kwargs: Any) -> None:
        """
        Perform logging without checking whether the logger is enabled for the level.

        Args:
            level: The logging level.
            *args: Additional positional arguments for the logger.
            **kwargs: Additional keyword arguments for the logger.

        Raises:
            ValueError: If no message is provided.
        """
        return super()._log(level, msg, args, **kwargs)

    def _log_fn_result(
        self,
        level: _LogLevel,
        callable: Callable[[T], str],
        *callable_args: T,
        **logging_kwargs: Any,
    ) -> None:
        """
        Call a function and log its result if the logger is enabled for the level.

        Args:
            level: The logging level.
            callable: A function that returns a string to log.
            *callable_args: Arguments to pass to the callable.
            **logging_kwargs: Additional keyword arguments for logging.
        """
        if self.isEnabledFor(level):
            return self._log_nocheck(level, callable(*callable_args), (), **logging_kwargs)

    # Daemon

    def _ensure_daemon(self) -> None:
        """
        Ensures that the stats daemon is running. If it's not running and conditions are met,
        it creates and starts the daemon task. If the daemon has finished, it raises any exception
        that occurred during its execution.

        Example:
            This method is automatically invoked when logger.enabled is accessed.
        """
        if (ENVS.COLLECT_STATS or self.enabled) and self._daemon is None:  # type: ignore [attr-defined,has-type]
            self._daemon = create_task(self._stats_daemon(), name="_StatsLogger debug daemon")
        elif self._daemon.done():
            raise self._daemon.exception()  # type: ignore [misc]

    async def _stats_daemon(self) -> None:
        """
        The main loop of the stats daemon. It continuously collects event loop times
        and periodically logs various statistics about the system's performance.

        Example:
            The daemon logs subprocess, brownie and event loop stats every 300 seconds.
        """
        start = time()
        time_since_notified = 0.0
        while True:
            await yield_to_loop()
            now = time()
            duration = now - start
            collector.event_loop_times.append(duration)
            start = time()
            time_since_notified += duration
            if time_since_notified > 300:
                self.log_subprocess_stats(level=logging.INFO)
                self.log_brownie_stats(level=logging.INFO)
                self.log_event_loop_stats(level=logging.INFO)
                time_since_notified = 0.0

    # TODO: MOVE COLLECTIONS UOT OF THIS CLASS

    def log_validation_error(self, method: RPCEndpoint, e: msgspec.ValidationError) -> None:
        """
        Log a validation error for a specific RPC method.

        Args:
            method: The RPC method that encountered the error.
            e: The validation error that occurred.

        Example:
            >>> _logger.log_validation_error("eth_call", some_validation_error)
        """
        enabled = self.isEnabledFor(DEVHINT)
        if COLLECT_STATS or enabled:
            collector.validation_errors[method].append(e)
        if enabled:
            self._log(
                DEVHINT,
                "ValidationError when decoding response for %s. This *should* not impact your script. If it does, you'll know.",
                (method,),
            )

    def log_types(self, method: RPCEndpoint, decoded: Any) -> None:
        """
        Log the types of decoded values for a specific RPC method.

        This method analyzes the decoded response data from an RPC call and logs
        information about the data types encountered. It's useful for understanding
        the structure of responses from different RPC methods.

        Args:
            method: The RPC method being logged.
            decoded: The decoded response data from the RPC call.

        Example:
            >>> _logger.log_types("eth_call", decoded_response)
        """
        # TODO fix this, use enabled check
        types = {type(v) for v in decoded.values()}
        self.devhint(f"my method and types: {method} {types}")
        if list in types:
            self._log_list_types(decoded.values())
        collector.types.update(types)

    def _log_list_types(self, values: Iterable[Any], level: _LogLevel = DEVHINT) -> None:
        """
        Log the types of items in a list.

        This internal method is used to analyze and log the types of elements
        found in list structures within RPC responses. It's particularly useful
        for understanding complex, nested data structures.

        Args:
            values: The list of values to analyze and log types for.
            level: The logging level to use. Defaults to DEVHINT.

        Example:
            >>> _logger._log_list_types(some_list)
        """
        list_types = {type(_) for v in values if isinstance(v, list) for _ in v}
        collector.types.update(list_types)
        if self.isEnabledFor(level):
            self._log(level, f"list types: {list_types}")


_Times = Deque[float]


@final
class _Collector:
    """Handles the collection and computation of stats-related data."""

    def __init__(self) -> None:
        self.errd_batches: Deque["JSONRPCBatch"] = deque(maxlen=500)
        """
        A deque that stores information about failed JSON-RPC batches.
        It has a maximum length of 500 to prevent unbounded memory usage.
        """

        self.durations: DefaultDict[str, _Times] = defaultdict(lambda: deque(maxlen=50_000))
        """
        A default dictionary that stores timing information for various operations.
        The keys are operation descriptors, and the values are deques of float values
        representing durations. Each deque has a maximum length of 50,000.
        """

        self.types: Set[Type[Any]] = set()
        """
        A set that stores all the types encountered during data collection.
        This is used for debugging and analysis purposes.
        """

        self.event_loop_times: _Times = deque(maxlen=50_000)
        """
        A deque that stores event loop execution times.
        It has a maximum length of 50,000 to limit memory usage.
        """

        # not implemented
        self.validation_errors: DefaultDict[RPCEndpoint, Deque[msgspec.ValidationError]] = (
            defaultdict(lambda: deque(maxlen=100))
        )
        """
        A default dictionary that stores validation errors encountered for each RPC endpoint.
        The keys are RPC endpoints, and the values are deques of ValidationError objects.
        Each deque has a maximum length of 100.
        """

    @property
    def avg_loop_time(self) -> float:
        """
        Calculates and returns the average event loop execution time.

        Returns:
            The average time taken for event loop iterations.

        Example:
            >>> avg_time = collector.avg_loop_time
        """
        return sum(collector.event_loop_times) / len(collector.event_loop_times)

    @property
    def count_active_brownie_calls(self) -> int:
        """
        Returns the number of currently active Brownie calls.

        Returns:
            The count of active Brownie calls.

        Example:
            >>> active_calls = collector.count_active_brownie_calls
        """
        return ENVS.BROWNIE_CALL_SEMAPHORE.default_value - ENVS.BROWNIE_CALL_SEMAPHORE.semaphore._value  # type: ignore [attr-defined, no-any-return]

    @property
    def count_queued_brownie_calls(self) -> int:
        """
        Returns the number of Brownie calls currently in the queue.

        Returns:
            The count of queued Brownie calls.

        Example:
            >>> queued_calls = collector.count_queued_brownie_calls
        """
        return len(ENVS.BROWNIE_CALL_SEMAPHORE.semaphore._waiters)  # type: ignore [attr-defined, no-any-return]

    @property
    def encoder_queue_len(self) -> int:
        """
        Returns the current length of the Brownie encoder queue.

        Returns:
            The number of items in the encoder queue.

        Example:
            >>> encoder_length = collector.encoder_queue_len
        """
        return ENVS.BROWNIE_ENCODER_PROCESSES._queue_count  # type: ignore [attr-defined, no-any-return]

    @property
    def decoder_queue_len(self) -> int:
        """
        Returns the current length of the Brownie decoder queue.

        Returns:
            The number of items in the decoder queue.

        Example:
            >>> decoder_length = collector.decoder_queue_len
        """
        return ENVS.BROWNIE_DECODER_PROCESSES._queue_count  # type: ignore [attr-defined, no-any-return]

    @property
    def mcall_decoder_queue_len(self) -> int:
        """
        Returns the current length of the multicall decoder queue.

        Returns:
            The number of items in the multicall decoder queue.

        Example:
            >>> mcall_length = collector.mcall_decoder_queue_len
        """
        return ENVS.MULTICALL_DECODER_PROCESSES._queue_count  # type: ignore [attr-defined, no-any-return]


@final
class _Writer:
    """
    Writer is used to turn Collector stats into human readable on a as-needed, JIT basis
    without wasting compute or cluttering Collector or StatsLogger class definitions.
    """

    def event_loop(self) -> str:
        """
        Generates a human-readable string describing the average event loop time.

        Returns:
            A string containing the average event loop time.

        Example:
            >>> print(_Writer().event_loop())
        """
        return f"Average event loop time: {collector.avg_loop_time}"

    def brownie(self) -> str:
        """
        Generates a human-readable string describing the current state of Brownie calls.

        Returns:
            A string containing the number of active and queued Brownie calls.

        Example:
            >>> print(_Writer().brownie())
        """
        return f"{collector.count_active_brownie_calls} brownie calls are processing, {collector.count_queued_brownie_calls} are queued in {ENVS.BROWNIE_CALL_SEMAPHORE}."

    def queue(self, pool: ProcessPoolExecutor) -> str:
        """
        Generates a human-readable string describing the state of a process pool's queue.

        Args:
            pool: The ProcessPoolExecutor to describe.

        Returns:
            A string containing the number of items in the pool's queue.

        Example:
            >>> print(_Writer().queue(some_pool))
        """
        return f"{pool} has {pool._queue_count} items in its queue"


@final
class _SentryExporter:
    """
    A class for exporting statistics and metrics from the :obj:`metrics` dict to Sentry.

    This class provides methods for setting tags and measurements in Sentry,
    which can be used for monitoring and debugging purposes.

    Each metric value will be fetched by calling `getattr(collector, metrics[k])`.
    If the result is a callable object, it will be called without args.

    See Also:
        :mod:`sentry_sdk`: The Sentry SDK for Python.
    """

    metrics: Final = {
        "active_eth_calls": "count_active_brownie_calls",
        "queued_eth_calls": "count_queued_brownie_calls",
        "encoder_queue": "encoder_queue_len",
        "decoder_queue": "decoder_queue_len",
        "loop_time": "avg_loop_time",
    }
    units: Final = {"loop_time": "seconds"}

    def push_measurements(self) -> None:
        """
        Push all metrics in `self.metrics` to Sentry.
        """
        if self._exc:
            raise deepcopy(self._exc).with_traceback(self._exc.__traceback__)
        for tag, attr_name in self.metrics.items():
            attr = getattr(collector, attr_name)
            if callable(attr):
                attr = attr()
            self.set_measurement(tag, attr, self.units.get(attr_name))  # type: ignore [misc]

    def push_envs(self) -> None:
        """
        Push environment variables set by the user to Sentry.

        Any exceptions during tagging are logged as warnings.

        Example:
            >>> sentry.push_envs()
        """
        for env, value in _ENVS.items():
            try:
                self.set_tag(env, value)  # type: ignore [call-arg, misc]
            except Exception as e:
                logger.warning(
                    f"Unable to set sentry tag {env} to {value}. See {e.__class__.__name__} below:"
                )
                logger.info(e, exc_info=True)

    set_tag: Final = _nocompile.set_tag
    """
    Set a tag for the current scope in Sentry.

    This is a reference to :func:`sentry_sdk.set_tag`.

    See Also:
        Sentry documentation on using tags:
        https://docs.sentry.io/platforms/python/enriching-events/tags/
    """

    set_measurement: Final = _nocompile.set_measurement
    """
    Set a measurement for the current scope in Sentry.

    This is a reference to :func:`sentry_sdk.set_measurement`.

    See Also:
        Sentry documentation on using measurements:
        https://docs.sentry.io/platforms/python/enriching-events/measurements/
    """

    _exc: Final = _nocompile.exc
    """
    Stores any ImportError that occurred when trying to import sentry_sdk.
    If this is not None, it indicates that Sentry integration is not available.
    """


logger: Final = _StatsLogger(__name__)
log: Final = logger.stats
devhint: Final = logger.devhint
collector: Final = _Collector()
sentry: Final = _SentryExporter()



from a_sync.functools import cached_property_unsafe as cached_property

from typing_extensions import Self


MODES = "default", "application", "infura"


class OperationMode(str):
    """
    Enum-like class that defines different operation modes for dank_mids.

    These modes affect how dank_mids optimizes requests and interacts
    with different Ethereum node providers.
    """

    @cached_property
    def application(self):
        """
        Checks if the current mode is 'application'.

        The 'application' mode keeps the event loop as unblocked as possible
        so an asyncio application can run as designed.
        """
        return self.mode == "application"

    @cached_property
    def default(self):
        """
        Checks if the current mode is 'default'.
        """
        return self.mode == "default"

    @cached_property
    def infura(self):
        """
        Checks if the current mode is 'infura'.

        The 'infura' mode minimizes the total number of calls sent to the node.
        """
        return self.mode == "infura"

    @property
    def mode(self) -> Self:
        """
        Retrieves the current operation mode.

        Raises:
            ValueError: If the mode is invalid.
        """
        if self not in MODES:
            raise ValueError(
                f"dank mids operation mode {self} is invalid", f"valid modes: {MODES}", str(self)
            )
        return self



from logging import getLogger
from typing import Final, Literal


batch_size_logger: Final = getLogger("dank_mids.batch_size")
_log_info: Final = batch_size_logger.info


def log_check(
    batch_type: Literal["multicall", "json"],
    member_type: Literal["calls", "requests"],
    num_calls: int,
) -> None:
    _log_info(
        "checking if we should reduce %s batch size... (%s %s)", batch_type, num_calls, member_type
    )



from typing import Any, Callable, Dict, Final, Generic, Iterable, Iterator, Optional, TypeVar, final
from weakref import ref


_T = TypeVar("_T")

GCCallback = Callable[[Any], None]


@final
class WeakList(Generic[_T]):
    def __init__(self, data: Optional[Iterable[_T]] = None) -> None:
        # Mapping from object ID to weak reference
        self._refs: Final[Dict[int, "ref[_T]"]] = {}
        if data is not None:
            self.extend(data)

    def __repr__(self) -> str:
        # Use list comprehension syntax within the repr function for clarity
        return f"{type(self).__name__}([{', '.join(map(repr, self))}])"

    def __len__(self) -> int:
        return len(self._refs)

    def __bool__(self) -> bool:
        return any(self)

    def __contains__(self, item: _T) -> bool:
        reference = self._refs.get(id(item))
        return False if reference is None else reference() is item

    def __iter__(self) -> Iterator[_T]:
        for reference in self._refs.values():
            obj = reference()
            if obj is not None:
                yield obj

    def append(self, item: _T) -> None:
        # Keep a weak reference with a callback for when the item is collected
        self._refs[id(item)] = ref(item, self._gc_callback)

    def extend(self, items: Iterable[_T]) -> None:
        callback = self._gc_callback
        self._refs.update((id(obj), ref(obj, callback)) for obj in items)

    def remove(self, item: _T) -> None:
        obj_id = id(item)
        reference = self._refs.get(obj_id)
        if reference is None or reference() is not item:
            raise ValueError("list.remove(x): x not in list")
        del self._refs[obj_id]

    def _gc_callback(self, item: Any) -> None:
        # Callback when a weakly-referenced object is garbage collected
        self._refs.pop(id(item), None)  # Safely remove the item if it exists



class Revert(ValueError): ...


class BrownieNotConnectedError(RuntimeError):
    """
    :class:`RuntimeError` raised when brownie is not connected but needs to be in order to access functionality within :mod:`dank_mids.brownie_patch`.
    """

    def __init__(self, obj_name: str):
        """
        Initializes a new BrownieNotConnectedError.

        Params:
            obj_name: The name of the object the user attempted to access, which requires brownie to be connected.
        """
        super().__init__(
            f"Brownie is not connected to a network. You must be connected to a network in order to access `dank_mids.{obj_name}`."
        )


class GarbageCollectionError(RuntimeError):
    """Exception raised when an object is garbage collected prematurely."""



from decimal import Decimal
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Optional,
    TypeVar,
)

from a_sync import igather
from brownie.typing import AccountsType
from eth_typing import ChecksumAddress
from hexbytes.main import BytesLike

from dank_mids import ENVIRONMENT_VARIABLES as ENVS
from dank_mids.brownie_patch import call
from dank_mids.brownie_patch._abi import FunctionABI


_EVMType = TypeVar("_EVMType")


class _DankMethodMixin(Generic[_EVMType]):
    """
    A mixin class that is used internally to enhance Brownie's contract methods
    with asynchronous support and memory optimization.
    """

    _address: ChecksumAddress
    """The address of the contract."""

    _abi: FunctionABI
    """The ABI (Application Binary Interface) of the contract function."""

    _prep_request_data: Callable[..., Awaitable[BytesLike]]

    def __init__(self) -> None:
        self._len_inputs = len(self.abi["inputs"])

        # cache `_call` on this object to prevent repeated imports due to circ import
        import dank_mids

        self._call = dank_mids.eth._call

        if ENVS.OPERATION_MODE.application or self._len_inputs:  # type: ignore [attr-defined]
            self._prep_request_data = call.encode
        else:
            self._prep_request_data = call._request_data_no_args

        self._skip_decoder_proc_pool = self._address in call._skip_proc_pool

    def __await__(self) -> Generator[Any, None, _EVMType]:
        """
        Allow the contract method to be awaited.

        This method enables using 'await' on the contract method, which will call
        the method without arguments at the latest block and return the result.
        """
        return self.coroutine().__await__()

    async def map(
        self,
        args: Iterable[Any],
        block_identifier: Optional[int] = None,
        *,
        iter_args: bool = False,
        decimals: Optional[int] = None,
    ) -> List[_EVMType]:
        """
        Asynchronously call the contract method with multiple sets of arguments.

        This method allows for efficient batch calling of the contract method
        with different arguments.

        Args:
            args: An iterable of argument sets to be passed to the method.
            block_identifier: The block number or identifier for the calls.
            iter_args: True if you are passing an Iterable of Iterables, False otherwise.
            decimals: The number of decimal places by which to scale numeric results.
        """
        if iter_args:
            return await igather(
                self.coroutine(*call_args, block_identifier=block_identifier, decimals=decimals)
                for call_args in args
            )
        else:
            return await igather(
                self.coroutine(call_arg, block_identifier=block_identifier, decimals=decimals)
                for call_arg in args
            )

    @property
    def abi(self) -> dict:
        """
        The ABI of the contract function.

        This property provides access to the complete ABI dictionary of the function.
        """
        return self._abi.abi

    @property
    def signature(self) -> str:
        """
        The function signature.

        This property returns the unique signature of the contract function,
        which is used to identify the function in transactions.
        """
        return self._abi.signature

    async def coroutine(
        self,
        *args: Any,
        block_identifier: Optional[int] = None,
        decimals: Optional[int] = None,
        override: Optional[Dict[str, str]] = None,
    ) -> _EVMType:
        """
        Asynchronously call the contract method with the given arguments.

        This method is the core of the asynchronous functionality while using dank_mids with eth_brownie.

        Args:
            *args: Variable length argument list for the contract method.
            block_identifier: The block number or identifier for the call.
            decimals: The number of decimals by which to scale numeric results.
            override: Optional parameters to override chain state in the call.

        Note:
            This is a generic method and should be implemented in subclasses.
        """
        raise NotImplementedError

    @property
    def _input_sig(self) -> str:
        return self._abi.input_sig


get_function_abi = FunctionABI.singleton


class _DankMethod(_DankMethodMixin):
    def __init__(
        self,
        address: ChecksumAddress,
        abi: Dict,
        name: str,
        owner: Optional[AccountsType],
        natspec: Optional[Dict] = None,
    ) -> None:
        self._address = address
        self._abi = get_function_abi(**abi)

        self._name = name
        """The name of the contract method."""

        self._owner = owner
        """The owner of the contract."""

        self.natspec = natspec or {}
        """The NatSpec documentation for the function."""

        super().__init__()

        self._encode_input = call.encode_input
        self._decode_output = call.decode_output

    async def coroutine(
        self,
        *args: Any,
        block_identifier: Optional[int] = None,
        decimals: Optional[int] = None,
        override: Optional[Dict[str, str]] = None,
    ) -> _EVMType:
        """
        Asynchronously call the contract method via dank mids and await the result.

        Args:
            *args: The arguments for the contract method.
            block_identifier: The block at which the chain will be read. If not provided, will read the chain at latest block.
            decimals: If provided, the output will be `result / 10 ** decimals`.

        Raises:
            ValueError: If the `override` parameter is used, as state overrides are not supported.

        Returns:
            The result of the contract method call.
        """
        if override:
            raise ValueError("Cannot use state override with `coroutine`.")

        encode_input_coro = self._encode_input(
            self, self._len_inputs, self._prep_request_data, *args
        )

        async with ENVS.BROWNIE_ENCODER_SEMAPHORE[block_identifier]:  # type: ignore [attr-defined,index]
            output_coro = self._call(  # type: ignore [call-arg]
                {"to": self._address, "data": await encode_input_coro},
                block_identifier,
            )
            async with ENVS.BROWNIE_CALL_SEMAPHORE[block_identifier]:  # type: ignore [attr-defined,index]
                decode_output_coro = self._decode_output(self, await output_coro)

        return (
            await decode_output_coro
            if decimals is None
            else await decode_output_coro / 10 ** Decimal(decimals)
        )



import logging
from typing import Final


getLogger: Final = logging.getLogger
CRITICAL: Final = logging.CRITICAL
FATAL: Final = logging.FATAL
ERROR: Final = logging.ERROR
WARN: Final = logging.WARN
WARNING: Final = logging.WARNING
INFO: Final = logging.INFO
DEBUG: Final = logging.DEBUG
NOTSET: Final = logging.NOTSET



import asyncio
from typing import (
    Coroutine,
    Final,
    Iterable,
    Literal,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import a_sync.asyncio


__T = TypeVar("__T")

FinishedTasks = Set["asyncio.Future[__T]"]
PendingTasks = Set["asyncio.Future[__T]"]


# These compile to C constants
Task: Final = asyncio.Task
get_running_loop: Final = asyncio.get_running_loop
wait: Final = asyncio.wait


yield_to_loop: Final = a_sync.asyncio.sleep0


async def gatherish(coros: Iterable[Coroutine], *, name: Optional[str] = None) -> None:  # type: ignore [type-arg]
    # sourcery skip: use-contextlib-suppress
    """
    An implementation of asyncio.gather optimized for use cases that:
        1. Have coroutines that are predomininately sync
        2. Expect occasional Exceptions but do not want to propagate them to the other tasks
        3. Do not need the results of the tasks returned"
    """
    loop = get_running_loop()

    # materialize the map into a list to make sure all the tasks start
    tasks = iter([Task(coro, loop=loop, name=name) for coro in coros])
    # sleep twice to let 99% of the tasks complete
    # NOTE: By doing this we allow any successful calls to get their results sooner without being interrupted
    #       by the first exc in the gather and having to wait for the bisect and retry process
    # TODO: stop retrying ones that succeed, that's wasteful
    await yield_to_loop()
    await yield_to_loop()
    for task in tasks:
        try:
            await task
        except Exception:
            # we will only retrieve the first exc from our tasks, if any
            # this hack prevents asyncio from logging a message that the other excs were not retrieved
            for task in tasks:
                # Make sure all the remaining tasks complete before we raise the exc
                try:
                    await task
                except Exception:
                    pass
            raise


@overload
async def first_completed(
    *fs: asyncio.Future[__T], cancel: Literal[True]
) -> FinishedTasks[__T]: ...
@overload
async def first_completed(
    *fs: asyncio.Future[__T], cancel: Literal[False] = False
) -> Tuple[FinishedTasks[__T], PendingTasks[__T]]: ...
async def first_completed(
    *fs: asyncio.Future[__T], cancel: bool = False
) -> Union[FinishedTasks[__T], Tuple[FinishedTasks[__T], PendingTasks[__T]]]:
    if not cancel:
        return await wait(fs, return_when="FIRST_COMPLETED")
    done, pending = await wait(fs, return_when="FIRST_COMPLETED")
    for p in pending:
        p.cancel()
    return done



from typing import Final

from dank_mids.helpers.lru_cache import lru_cache_lite_nonull


BYPASS_METHODS: Final = "eth_blockNumber", "eth_getLogs", "trace_", "debug_"
"""
A tuple of method names that should bypass batching.
These methods are typically handled separately or have special requirements.
"""


@lru_cache_lite_nonull
def get_len(method: str) -> int:
    # NOTE: These are totally arbitrary, used to reduce frequency of giant batches/responses
    if method == "eth_getTransactionReceipt":
        return 5
    elif method in {"eth_getTransaction", "eth_getCode"} or "eth_getBlockBy" in method:
        return 3
    return 1


@lru_cache_lite_nonull
def should_batch(method: str) -> bool:
    return all(bypass not in method for bypass in BYPASS_METHODS)



from typing import Any, Dict, Optional, Tuple, TypeVar, Union, final

from brownie.network.contract import ContractCall, ContractTx, OverloadedMethod
from brownie.typing import AccountsType
from eth_typing import ChecksumAddress

from dank_mids.brownie_patch._method import _DankMethod, _DankMethodMixin, _EVMType


_T = TypeVar("_T")

ContractMethod = Union[ContractCall, ContractTx, OverloadedMethod]
"""Type alias for Brownie contract methods."""


@final
class DankContractCall(_DankMethod, ContractCall):  # type: ignore [misc]
    """
    A subclass of `brownie.network.contract.ContractCall` with async support via the `coroutine` method.

    This class uses less memory than a typical `ContractTx` by employing the `__slots__` functionality alongside
    the `FunctionABI` singleton to manage the function ABI and related logic.

    You can await an instance of this class directly to call the contract method without arguments at the latest block.
    """


@final
class DankContractTx(_DankMethod, ContractTx):  # type: ignore [misc]
    """
    A subclass of `brownie.network.contract.ContractTx` with async support via the `coroutine` method.

    This class is optimized for memory use compared to a standard `ContractTx`, by using `__slots__` in combination
    with the `FunctionABI` singleton to handle the function ABI and associated logic.

    Awaiting an instance of this class directly executes the contract method without arguments at the latest block.
    """


_NonOverloaded = Union[DankContractCall, DankContractTx]
"""Type alias for non-overloaded Dank contract methods."""


@final
class DankOverloadedMethod(OverloadedMethod, _DankMethodMixin[_T]):  # type: ignore [misc]
    """
    A subclass of `brownie.network.contract.OverloadedMethod` with async support via the `coroutine` method.

    This class is designed to be memory efficient compared to a typical `OverloadedMethod`, employing `__slots__`
    and the `FunctionABI` singleton to manage the function ABI and related logic.

    You can await instances of this class directly to call the contract method without arguments at the latest block.
    """

    methods: Dict[Tuple[str, ...], _NonOverloaded]

    async def coroutine(
        self,
        *args: Any,
        block_identifier: Optional[int] = None,
        decimals: Optional[int] = None,
        override: Optional[Dict[str, str]] = None,
    ) -> _EVMType:
        """
        Asynchronously call the contract method using dank mids and await the result.

        Args:
            *args: The arguments for the contract method.
            block_identifier (optional): Specifies the block at which the chain will be read. Defaults to the latest block if not provided.
            decimals (optional): If specified, the output will be `result / 10 ** decimals`.

        Returns:
            Whatever the node sends back as the output for this contract method.
        """
        call: Union[DankContractCall, DankContractTx] = self._get_fn_from_args(args)
        return await call.coroutine(
            *args, block_identifier=block_identifier, decimals=decimals, override=override
        )

    def _add_fn(self, abi: Dict, natspec: Dict) -> None:  # type: ignore [type-arg]
        """
        Add a function to the overloaded method.

        This method creates a new function object using the provided ABI and NatSpec,
        then adds it to the collection of functions in the overloaded method.

        Args:
            abi: The ABI of the function to be added.
            natspec: The NatSpec documentation for the function.
        """
        fn = _get_method_object(self._address, abi, self._name, self._owner, natspec)
        key = tuple(i["type"].replace("256", "") for i in abi["inputs"])
        self.methods[key] = fn
        self.natspec.update(natspec)


DankContractMethod = Union[DankContractCall, DankContractTx, DankOverloadedMethod]  # type: ignore [type-arg]
"""
Alias for `ContractMethod` objects with async support via an additional `coroutine` method.

These objects consume less memory than standard `ContractMethod` objects by utilizing the `FunctionABI` singleton to handle the function ABI and associated logic.

Instances of this class can be awaited directly to call the contract method without arguments at the latest block.
"""


def _get_method_object(
    address: ChecksumAddress,
    abi: Dict[str, Any],
    name: str,
    owner: Optional[AccountsType],
    natspec: Dict[str, Any],
) -> Union[DankContractCall, DankContractTx]:
    # sourcery skip: default-get
    """
    Retrieve an appropriate method object based on the ABI.

    Determines whether the ABI specifies a constant function and returns a corresponding DankContractCall
    or DankContractTx object.

    Args:
        address: The contract address associated with the method.
        abi: The ABI of the function.
        name: The name of the method.
        owner: The account that owns the method, if applicable.
        natspec: The NatSpec documentation associated with the function.

    Returns:
        A DankContractCall or DankContractTx object depending on the function's mutability state.
    """
    if "constant" in abi:
        constant = abi["constant"]
    else:
        constant = abi["stateMutability"] in ("view", "pure")

    if constant:
        return DankContractCall(address, abi, name, owner, natspec)
    return DankContractTx(address, abi, name, owner, natspec)



"""
This helper file lets us conditionally define some imports,
based on whether or not the user has `sentry_sdk` installed.

Mypyc does not currently handle conditional imports like this well.
"""

from typing import Any, Callable, Optional


set_tag: Optional[Callable[[str, Any], None]]
"""
Set a tag for the current scope in Sentry.

This is a reference to :func:`sentry_sdk.set_tag`.

See Also:
    Sentry documentation on using tags:
    https://docs.sentry.io/platforms/python/enriching-events/tags/
"""


set_measurement: Optional[Callable]
"""
Set a measurement for the current scope in Sentry.

This is a reference to :func:`sentry_sdk.set_measurement`.

See Also:
    Sentry documentation on using measurements:
    https://docs.sentry.io/platforms/python/enriching-events/measurements/
"""


exc: Optional[ImportError]
"""
Stores any ImportError that occurred when trying to import sentry_sdk.
If this is not None, it indicates that Sentry integration is not available.
"""


try:
    import sentry_sdk

    set_tag = sentry_sdk.set_tag
    set_measurement = sentry_sdk.set_measurement
    exc = None
except ImportError as e:
    exc = e
    set_tag = None
    set_measurement = None



from dank_mids.helpers._gather import gatherish
from dank_mids.helpers._helpers import setup_dank_w3, setup_dank_w3_from_sync
from dank_mids.helpers.future import DebuggableFuture
from dank_mids.helpers.hashing import make_hashable
from dank_mids.helpers.lru_cache import lru_cache_lite, lru_cache_lite_nonull

__all__ = [
    "setup_dank_w3",
    "setup_dank_w3_from_sync",
    "DebuggableFuture",
    "lru_cache_lite",
    "lru_cache_lite_nonull",
    "make_hashable",
    "gatherish",
]



from typing import Final

import a_sync
from typed_envs import EnvVarFactory

from dank_mids.semaphores import BlockSemaphore


_factory: Final = EnvVarFactory("DANKMIDS")

# TODO: maybe have ez-a-sync do this natively for its own classes?
# TODO: maybe have typed-envs do this natively for builtin types? (asyncio.Semaphore, cf.Executor, etc...)
_factory.register_string_converter(a_sync.Semaphore, int)
_factory.register_string_converter(a_sync.ProcessPoolExecutor, int)
_factory.register_string_converter(BlockSemaphore, int)

create_env: Final = _factory.create_env

# This only applies to the oldest of dank's envs
_deprecated_format: Final = EnvVarFactory("DANK_MIDS")



import http
from asyncio import CancelledError, Task, create_task, current_task, sleep
from collections import defaultdict
from enum import IntEnum
from heapq import nlargest
from itertools import chain
from random import random
from threading import get_ident
from time import time
from typing import Any, Callable, DefaultDict, Dict, Final, Tuple, final, overload

from a_sync import Event
from a_sync._smart import shield
from a_sync.asyncio import sleep0 as yield_to_loop
from aiohttp import ClientSession, ClientTimeout, TCPConnector
from aiohttp.client_exceptions import ClientResponseError
from aiohttp.typedefs import DEFAULT_JSON_DECODER, JSONDecoder
from aiolimiter import AsyncLimiter
from async_lru import alru_cache

from dank_mids import ENVIRONMENT_VARIABLES as ENVS
from dank_mids._logging import DEBUG, getLogger
from dank_mids.helpers._codec import JSONRPCBatchResponse, RawResponse
from dank_mids.types import PartialRequest


logger: Final = getLogger("dank_mids.session")


# NOTE: You cannot subclass an IntEnum object so we have to do some hacky shit here.
# First, set up custom error codes we might see.
class _HTTPStatusExtension(IntEnum):
    """
    A modified version of :class:`http.HTTPStatus`, for custom codes used by specific services like Cloudflare.

    This enum contains status codes and descriptions for server errors that are not part of the standard HTTP specification.
    """

    WEB_SERVER_IS_RETURNING_AN_UNKNOWN_ERROR = (
        520,
        "Web Server is Returning an Unknown Error",
        "HTTP response status code 520 Web server is returning an unknown error is an unofficial server error\n"
        + "that is specific to Cloudflare. This is a catch-all error that is used in the absence of having a\n"
        + "HTTP status code for one that is more specific.\n"
        + "Learn more at https://http.dev/520",
    )

    WEB_SERVER_IS_DOWN = (
        521,
        "Web Server Is Down",
        "HTTP response status code 521 Web server is down is an unofficial server error that is specific to Cloudflare.\n"
        + "This HTTP status code occurs when the HTTP client was able to successfully connect to Cloudflare but it was\n"
        + "unable to connect to the origin server."
        + "Learn more at https://http.dev/521",
    )

    CLOUDFLARE_CONNECTION_TIMEOUT = (
        522,
        "Cloudflare Connection Timeout",
        "Cloudflare is a content delivery network that acts as a gateway between a user and a website server.\n"
        + "When the 522 Connection timed out status code is received, Cloudflare attempted to connect\n"
        + "to the origin server but was not successful due to a timeout. The HTTP Connection was not established\n"
        + "most likely because the IP addresses of Cloudflare are blocked by the origin server,\n"
        + "the origin server settings are misconfigured, or the origin server is overloaded."
        + "Learn more at https://http.dev/522",
    )

    CLOUDFLARE_TIMEOUT = (
        524,
        "Cloudflare Timeout",
        '"524 A timeout occurred" is an unofficial server error that is specific to Cloudflare.\n'
        + "When the 524 A timeout occurred status code is received, it implies that a successful\n"
        + "HTTP Connection was made between Cloudflare and the origin server, however the connection\n"
        + "timed out before the HTTP request was completed. Cloudflare typically waits 100 seconds\n"
        + "for an HTTP response and returns this HTTP status code if nothing is received.\n"
        + "Learn more at https://http.dev/524",
    )

    SITE_FROZEN = (
        530,
        "Site Frozen",
        "HTTP response status code 530 is an unofficial server error that is specific to Cloudflare\n"
        + "and Pantheon. In the case of Cloudflare, a second HTTP status code 1XXX error message will be\n"
        + "included to provide a more accurate description of the problem. For Pantheon, HTTP status code\n"
        + "530 Site Frozen indicates that a site has been restricted due to inactivity.\n"
        + "Learn more at https://http.dev/530",
    )

    def __new__(cls, value, phrase, description=""):
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.phrase = phrase  # type: ignore [attr-defined]
        obj.description = description  # type: ignore [attr-defined]
        return obj


# Then, combine the standard HTTPStatus enum and the custom extension to get the full custom HTTPStatus enum we need.
HTTPStatusExtended: Final = IntEnum("HTTPStatusExtended", [(i.name, i.value) for i in chain(http.HTTPStatus, _HTTPStatusExtension)])  # type: ignore [misc]
"""
An extension of :class:`http.HTTPStatus`, supporting both standard HTTP status codes as well as custom codes used by specific services like Cloudflare.

This enum includes status codes and descriptions for server errors that are not part of the standard HTTP specification.
"""

RETRY_FOR_CODES: Final = {
    HTTPStatusExtended.BAD_GATEWAY,  # type: ignore [attr-defined]
    HTTPStatusExtended.WEB_SERVER_IS_RETURNING_AN_UNKNOWN_ERROR,  # type: ignore [attr-defined]
    HTTPStatusExtended.CLOUDFLARE_CONNECTION_TIMEOUT,  # type: ignore [attr-defined]
    HTTPStatusExtended.CLOUDFLARE_TIMEOUT,  # type: ignore [attr-defined]
}


def _get_status_enum(error: ClientResponseError) -> HTTPStatusExtended:
    try:
        return HTTPStatusExtended(error.status)
    except ValueError as ve:
        if str(ve).endswith("is not a valid HTTPStatusExtended"):
            # we still want the original exc to raise
            raise error from ve
        raise


# default is 50 requests/second
limiters: Final[DefaultDict[str, AsyncLimiter]] = defaultdict(
    lambda: AsyncLimiter(1, 1 / ENVS.REQUESTS_PER_SECOND)  # type: ignore [operator]
)

_rate_limit_waiters: Final[Dict[str, Event]] = {}


async def rate_limit_inactive(endpoint: str) -> None:
    """
    Wait until the rate limiter for `endpoint` has no remaining waiters.
    If someone's already waiting on this endpoint, just await their Event.
    Otherwise, set up our own Event and signal once no waiters remain.
    """
    # Quick exit if no queued waiters
    if not limiters[endpoint]._waiters:
        return

    # If another task is already waiting, just join it
    if existing := _rate_limit_waiters.get(endpoint):
        await existing.wait()
        return

    # Otherwise, create an Event for others to wait on
    task = _rate_limit_tasks.get(endpoint)
    if task is None:
        _rate_limit_waiters[endpoint] = Event(endpoint, 10)
        task = _rate_limit_tasks[endpoint] = create_task(__rate_limit_inactive(endpoint))
    await shield(task)


_rate_limit_tasks: Final[Dict[str, "Task[None]"]] = {}


async def __rate_limit_inactive(endpoint: str) -> None:
    # sourcery skip: use-contextlib-suppress
    waiters = limiters[endpoint]._waiters
    while waiters:
        # pop last item
        last_waiter_tuple = nlargest(1, waiters)[0]
        last_waiter = last_waiter_tuple[-1]

        if last_waiter.cancelled():
            waiters.remove(last_waiter_tuple)
            continue

        if last_waiter.done():
            # NOTE: I don't think this is possible but want to confirm
            _rate_limit_waiters.pop(endpoint).set()
            _rate_limit_tasks.pop(endpoint)
            raise RuntimeError("last waiter is done")

        # await it
        try:
            await last_waiter
        except CancelledError:
            # AsyncLimiter cancels the fut as part of regular operation
            pass

        # let recently popped waiters check the limiter for capacity, they might create new waiters
        # then, let recently popped waiters make some calls to see if we're still being limited
        for _ in range(10):
            if waiters:
                break
            await yield_to_loop()

    _rate_limit_waiters.pop(endpoint).set()
    _rate_limit_tasks.pop(endpoint)


@overload
async def post(
    endpoint: str, *args, loads: Callable[[Any], RawResponse], **kwargs
) -> RawResponse: ...
@overload
async def post(
    endpoint: str, *args, loads: Callable[[Any], JSONRPCBatchResponse], **kwargs
) -> JSONRPCBatchResponse: ...
async def post(endpoint: str, *args, loads: JSONDecoder = DEFAULT_JSON_DECODER, **kwargs) -> Any:
    """Returns decoded json data from `endpoint`"""
    session = await get_session()
    return await session.post(endpoint, *args, loads=loads, **kwargs)


async def get_session() -> "DankClientSession":
    return await _get_session_for_thread(get_ident())


@final
class DankClientSession(ClientSession):
    _limited = False
    _last_rate_limited_at = 0
    _continue_requests_at = 0

    async def post(self, endpoint: str, *args, loads: JSONDecoder = DEFAULT_JSON_DECODER, **kwargs) -> bytes:  # type: ignore [override]
        if (now := time()) < self._continue_requests_at:
            await sleep(self._continue_requests_at - now)

        # Process input arguments.
        data = kwargs.get("data")
        if debug_logs_enabled := _logger_is_enabled_for(DEBUG):
            if isinstance(data, PartialRequest):
                kwargs["data"] = data.data
                _logger_log(DEBUG, "making request for %s", (data,))
            _logger_log(
                DEBUG,
                "making request to %s with (args, kwargs): (%s %s)",
                (endpoint, args, kwargs),
            )
        elif isinstance(data, PartialRequest):
            kwargs["data"] = data.data

        # Try the request until success or 5 failures.
        tried = 0
        while True:
            try:
                async with limiters[endpoint]:
                    async with ClientSession.post(self, endpoint, *args, **kwargs) as response:
                        response_data = await response.json(loads=loads, content_type=None)
                        _logger_debug("received response %s", response_data)
                        return response_data
            except ClientResponseError as ce:
                status = ce.status
                if status == HTTPStatusExtended.TOO_MANY_REQUESTS:  # type: ignore [attr-defined]
                    await self._handle_too_many_requests(endpoint, ce)

                elif status in RETRY_FOR_CODES and tried < 5:
                    tried += 1
                    if debug_logs_enabled:
                        sleep_for = random()
                        _logger_log(
                            DEBUG,
                            "response failed with status %s, retrying in %.f2s",
                            (HTTPStatusExtended(status), sleep_for),
                        )
                        await sleep(sleep_for)
                    else:
                        await sleep(random())

                else:
                    if debug_logs_enabled:
                        _logger_log(
                            DEBUG,
                            "response failed with status %s  request data: %s",
                            (_get_status_enum(ce), data),
                        )
                    raise

    async def _handle_too_many_requests(self, endpoint: str, error: ClientResponseError) -> None:
        limiter = limiters[endpoint]
        if (now := time()) > getattr(limiter, "_last_updated_at", 0) + 10:
            current_rate = limiter._rate_per_sec
            new_rate = current_rate * 0.97
            if new_rate >= ENVS.MIN_REQUESTS_PER_SECOND:
                limiter.time_period /= 0.97
                limiter._rate_per_sec = new_rate
                limiter._last_updated_at = now
                _logger_info(
                    "reduced requests per second for %s from %s to %s",
                    endpoint,
                    round(current_rate, 3),
                    round(new_rate, 3),
                )

        now = time()
        self._last_rate_limited_at = now
        secs_between_requests = 1 / limiter._rate_per_sec
        retry_after = float(error.headers.get("Retry-After", secs_between_requests * 5))
        resume_at = max(
            self._continue_requests_at + retry_after,
            self._last_rate_limited_at + retry_after,
        )
        retry_after = resume_at - now
        self._continue_requests_at = resume_at

        self._log_rate_limited(retry_after)
        if retry_after > 30:
            _logger_warning("severe rate limiting from your provider")
        # the limiter handles the timing
        limiter._level += 5

    def _log_rate_limited(self, try_after: float) -> None:
        if not self._limited:
            self._limited = True
            _logger_info("You're being rate limited by your node provider")
            _logger_info(
                "Its all good, dank_mids has this handled, but you might get results slower than you'd like"
            )
        if try_after < 5:
            _logger_debug("rate limited: retrying after %.3fs", try_after)
        else:
            _logger_info("rate limited: retrying after %.3fs", try_after)


@alru_cache(maxsize=None)
async def _get_session_for_thread(thread_ident: int) -> DankClientSession:
    """
    This makes our ClientSession threadsafe just in case.
    Most everything should be run in main thread though.
    """
    # I'm testing the value to use for limit, eventually will make an env var for this with an appropriate default
    connector = TCPConnector(limit=0, enable_cleanup_closed=True)
    client_timeout = ClientTimeout(  # type: ignore [arg-type, attr-defined]
        int(ENVS.AIOHTTP_TIMEOUT)
    )
    return DankClientSession(
        connector=connector,
        headers={"content-type": "application/json"},
        timeout=client_timeout,
        raise_for_status=True,
        read_bufsize=2**20,  # 1mb
    )


def _logger_is_enabled_for(level: int) -> bool: ...
def _logger_warning(msg: str, *args: Any) -> None: ...
def _logger_info(msg: str, *args: Any) -> None: ...
def _logger_debug(msg: str, *args: Any) -> None: ...
def _logger_log(level: int, msg: str, args: Tuple[Any, ...]) -> None: ...


_logger_is_enabled_for: Final = logger.isEnabledFor
_logger_warning: Final = logger.warning
_logger_info: Final = logger.info
_logger_debug: Final = logger.debug
_logger_log: Final = logger._log



from functools import wraps
from types import MethodType
from typing import Any, Dict, Optional, Union

from brownie import Contract
from brownie.network.contract import ContractCall, ContractTx, OverloadedMethod
from dank_mids.brownie_patch.call import _get_coroutine_fn, _skip_proc_pool
from dank_mids.helpers._helpers import DankWeb3


def _patch_overloaded_method(call: OverloadedMethod, w3: DankWeb3) -> None:
    """
    Patch an overloaded contract method with async functionality.

    Args:
        call: The overloaded method to patch. Currently, only methods of type
              OverloadedMethod are fully supported.
        w3: The DankWeb3 instance.
    """

    # sourcery skip: avoid-builtin-shadow
    @wraps(call)
    async def coroutine(
        self: Contract,
        *args: Any,
        block_identifier: Optional[Union[int, str, bytes]] = None,
        decimals: Optional[int] = None,
        override: Optional[Dict[str, str]] = None,
    ) -> Any:
        """
        Coroutine to call the overloaded contract method.

        Args:
            self: The contract instance.
            *args: Positional arguments to pass to the contract function.
            block_identifier: Optional block identifier to query state at a specific block.
            decimals: Optional number of decimals for the return value.
            override: Optional dictionary of overrides.

        Raises:
            ValueError: If the function cannot be determined from the arguments.
        """
        try:
            fn = self._get_fn_from_args(args)
        except ValueError as e:
            exc_str = str(e)
            match_on = (
                f"Contract has more than one function '.{call._name}",
                f"You must explicitly declare which function you are calling, e.g. .{call._name}",
            )
            if all(s in exc_str for s in match_on):
                breakpoint = exc_str.find("(*args)")
                raise ValueError(f"{exc_str[:breakpoint]}.coroutine{exc_str[breakpoint:]}")
            raise

        kwargs = {"block_identifier": block_identifier, "decimals": decimals, "override": override}
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return await fn.coroutine(*args, **kwargs)

    for method in call.__dict__["methods"].values():
        if isinstance(method, (ContractCall, ContractTx)):
            method._skip_decoder_proc_pool = method._address in _skip_proc_pool  # type: ignore [union-attr]
            method.coroutine = MethodType(_get_coroutine_fn(w3, len(method.abi["inputs"])), method)  # type: ignore [union-attr]
    # TODO implement this properly
    # elif isinstance(call, ContractTx):
    # _patch_tx(call, w3)

    call.coroutine = MethodType(coroutine, call)  # type: ignore [attr-defined]



from dank_mids.helpers._lock import AlertingRLock


class UIDGenerator:
    __slots__ = "_value", "_lock"

    def __init__(self) -> None:
        """
        Initializes the UIDGenerator with the starting UID value set to -1 and a custom reentrant lock.

        This lock is used to ensure thread-safety during UID generation.

        Examples:
            >>> uid_gen = UIDGenerator()
            >>> uid_gen.latest
            -1
            >>> uid_gen.next
            0

        See Also:
            :class:`AlertingRLock`
        """
        self._value: int = -1
        self._lock = AlertingRLock(name="uid")

    @property
    def latest(self) -> int:
        """Retrieves the most recent unique ID without incrementing.

        Examples:
            >>> uid_gen = UIDGenerator()
            >>> uid_gen.latest
            -1
        """
        return self._value

    @property
    def next(self) -> int:
        """Generates and retrieves the next unique ID in a thread-safe manner.

        Examples:
            >>> uid_gen = UIDGenerator()
            >>> uid_gen.next
            0
            >>> uid_gen.latest
            0
        """
        with self._lock:
            new = self._value = self._value + 1
            return new



import decimal
from concurrent.futures.process import BrokenProcessPool
from logging import Logger
from pickle import PicklingError
from types import MethodType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Final,
    List,
    NewType,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import brownie.convert.datatypes
import brownie.convert.normalize
import brownie.network.contract
import eth_abi
import hexbytes
from a_sync import AsyncProcessPoolExecutor
from brownie import chain
from brownie.convert.normalize import ABIType
from brownie.convert.utils import get_type_strings
from brownie.exceptions import VirtualMachineError
from brownie.network.contract import ContractCall
from brownie.project.compiler.solidity import SOLIDITY_ERROR_CODES
from eth_abi.exceptions import DecodingError, InsufficientDataBytes
from eth_typing import HexStr
from evmspec.data import Address
from hexbytes.main import BytesLike
from multicall.constants import MULTICALL2_ADDRESSES
from web3.types import BlockIdentifier

from dank_mids import ENVIRONMENT_VARIABLES as ENVS
from dank_mids import exceptions
from dank_mids._logging import getLogger
from dank_mids.helpers.lru_cache import lru_cache_lite_nonull
from dank_mids.helpers._helpers import DankWeb3

if TYPE_CHECKING:
    from dank_mids.brownie_patch.types import ContractMethod


APPLICATION_MODE: Final[bool] = ENVS.OPERATION_MODE.application


_T = TypeVar("_T")
TypeStr = NewType("TypeStr", str)
TypeStrs = List[TypeStr]
ListOrTuple = Union[List[_T], Tuple[_T, ...]]
AbiDict = NewType("AbiDict", Dict[str, Any])


# these compile to C constants to avoid global name lookups
Decimal: Final[Callable[[Any], decimal.Decimal]] = decimal.Decimal
HexBytes: Final = hexbytes.HexBytes
Contract: Final = brownie.network.contract.Contract
ReturnValue: Final = brownie.convert.datatypes.ReturnValue
HexString: Final = brownie.convert.normalize.HexString
TupleType: Final = brownie.convert.normalize.TupleType

Revert: Final = exceptions.Revert

to_bool: Final = brownie.convert.normalize.to_bool
to_decimal: Final = brownie.convert.normalize.to_decimal
to_int: Final = brownie.convert.normalize.to_int
to_string: Final = brownie.convert.normalize.to_string
to_uint: Final = brownie.convert.normalize.to_uint
_check_array: Final = brownie.convert.normalize._check_array
_get_abi_types: Final = brownie.convert.normalize._get_abi_types


logger: Final[Logger] = getLogger(__name__)


encode: Final = lambda self, *args: ENVS.BROWNIE_ENCODER_PROCESSES.run(__encode_input, self.abi, self.signature, *args)  # type: ignore [attr-defined]
"""
A lambda function that encodes input data for contract calls.

This function uses the Brownie encoder process pool (BROWNIE_ENCODER_PROCESSES) to run the
synchronous function :func:`__encode_input` in a separate process asynchronously.

Args:
    self: The contract method instance.
    *args: The arguments to be encoded.

Examples:
    >>> encoded = encode(contract_call_instance, arg1, arg2)
    >>> # encoded is a hex string that represents the ABI-encoded function call.

See Also:
    :func:`__encode_input`
"""

decode: Final[Callable[[Any, bytes], Any]] = lambda self, data: ENVS.BROWNIE_DECODER_PROCESSES.run(__decode_output, data, self.abi)  # type: ignore [attr-defined]
"""
A lambda function that decodes output data for contract calls.

This function uses the Brownie decoder process pool (BROWNIE_DECODER_PROCESSES) to run the
synchronous function :func:`__decode_output` asynchronously.

Args:
    self: The contract method instance.
    data: The raw byte data returned from a contract call.

Examples:
    >>> decoded = decode(contract_call_instance, output_data)
    >>> # decoded now holds the parsed return value from the contract call.

See Also:
    :func:`__decode_output`
"""

# We assign this variable so ypricemagic's checksum cache monkey patch will work,
# This is only relevant to you if your project uses ypricemagic as well.
to_checksum_address: Final = Address.checksum


def _patch_call(call: ContractCall, w3: DankWeb3) -> None:
    """
    Patch a Brownie ContractCall to enable asynchronous use via dank_mids for batching.

    Args:
        call: The original Brownie ContractCall to be patched.
        w3: An instance of DankWeb3 used to create the coroutine.

    The function modifies the ContractCall in place to add asynchronous capabilities.
    """
    call._skip_decoder_proc_pool = call._address in _skip_proc_pool
    call.coroutine = MethodType(_get_coroutine_fn(w3, len(call.abi["inputs"])), call)
    call.__await__ = MethodType(_call_no_args, call)


@lru_cache_lite_nonull
def _get_coroutine_fn(w3: DankWeb3, len_inputs: int) -> Callable[..., Any]:
    if APPLICATION_MODE or len_inputs:
        get_request_data = encode
    else:
        get_request_data = _request_data_no_args  # type: ignore [assignment]

    async def coroutine(
        self: ContractCall,
        *args: Any,
        block_identifier: Optional[BlockIdentifier] = None,
        decimals: Optional[int] = None,
        override: Optional[Dict[str, str]] = None,
    ) -> Any:
        if override:
            raise ValueError("Cannot use state override with `coroutine`.")
        async with ENVS.BROWNIE_ENCODER_SEMAPHORE[block_identifier]:  # type: ignore [attr-defined, index]
            data = await encode_input(self, len_inputs, get_request_data, *args)
            async with ENVS.BROWNIE_CALL_SEMAPHORE[block_identifier]:  # type: ignore [attr-defined, index]
                output = await w3.eth.call({"to": self._address, "data": data}, block_identifier)
        try:
            decoded = await decode_output(self, output)
        except InsufficientDataBytes as e:
            raise InsufficientDataBytes(
                {
                    "error": f"{e} when decoding response",
                    "method": self,
                    "contract": self._address,
                    "args": args,
                    "block": block_identifier,
                    "response": output,
                }
            ) from e

        return decoded if decimals is None else decoded / 10 ** Decimal(decimals)

    return coroutine


def _call_no_args(self: "ContractMethod") -> Any:
    """Asynchronously call `self` with no arguments at the latest block."""
    return self.coroutine().__await__()


async def encode_input(call: ContractCall, len_inputs: int, get_request_data: Callable, *args: Any) -> HexStr:  # type: ignore [type-arg]
    # We will just assume containers contain a Contract object until we have a better way to handle this
    if any(isinstance(arg, Contract) or hasattr(arg, "__contains__") for arg in args):
        # We can't unpickle these because of the added `coroutine` method.
        data = __encode_input(call.abi, call.signature, *args)
    else:
        try:  # We're better off sending these to the subprocess so they don't clog up the event loop.
            data = await get_request_data(call, *args)
        except (AttributeError, TypeError):
            # These occur when we have issues pickling an object, but that's fine, we can do it sync.
            data = __encode_input(call.abi, call.signature, *args)
        # TODO: move this somewhere else
        except BrokenProcessPool:
            logger.critical("Oh fuck, you broke the %s while decoding %s with abi %s", ENVS.BROWNIE_ENCODER_PROCESSES, data, call.abi)  # type: ignore [attr-defined]
            # Let's fix that right up
            ENVS.BROWNIE_ENCODER_PROCESSES = AsyncProcessPoolExecutor(ENVS.BROWNIE_ENCODER_PROCESSES._max_workers)  # type: ignore [attr-defined, assignment]
            data = __encode_input(call.abi, call.signature, *args) if len_inputs else call.signature
        except PicklingError:  # But if that fails, don't worry. I got you.
            data = __encode_input(call.abi, call.signature, *args) if len_inputs else call.signature
    # We have to do it like this so we don't break the process pool.
    if isinstance(data, Exception):
        if len(data.args) != 1:
            raise data
        exc_str = str(data)
        if exc_str.startswith(call._name):
            exc_str = exc_str.split(" ", 1)[1]
        else:
            exc_str = f": {exc_str}"
        raise type(data)(f"Contract('{call._address}').{call._name}: {exc_str}") from None
    return data


async def decode_output(call: ContractCall, data: bytes) -> Any:
    __validate_output(call.abi, data)
    try:
        if call._skip_decoder_proc_pool or b"Unexpected error" in data:  # Multicall3
            # This will break the process pool
            decoded = __decode_output(data, call.abi)
        else:
            try:
                decoded = await decode(call, data)
            # TODO: move this somewhere else
            except BrokenProcessPool:
                # Let's fix that right up
                logger.critical("Oh fuck, you broke the %s while decoding %s with abi %s", ENVS.BROWNIE_DECODER_PROCESSES, data, call.abi)  # type: ignore [attr-defined]
                ENVS.BROWNIE_DECODER_PROCESSES = AsyncProcessPoolExecutor(ENVS.BROWNIE_DECODER_PROCESSES._max_workers)  # type: ignore [attr-defined, assignment]
                decoded = __decode_output(data, call.abi)
        # We have to do it like this so we don't break the process pool.
        if isinstance(decoded, Exception):
            raise decoded
        return decoded
    except InsufficientDataBytes as e:
        # Add some context to the exception for the sake of the end-user.
        e.args = *e.args, call, call._address, data
        raise
    except AttributeError as e:
        # NOTE: Not sure why this happens as we set the attr while patching the call but w/e, this works for now.
        if not str(e).endswith(" object has no attribute '_skip_decoder_proc_pool'"):
            raise
        logger.debug("DEBUG ME BRO: %s", e)
        call._skip_decoder_proc_pool = call._address in _skip_proc_pool
        return await decode_output(call, data)


async def _request_data_no_args(call: ContractCall) -> HexStr:
    return call.signature  # type: ignore [return-value, no-any-return]


# These methods were renamed in eth-abi 4.0.0
__eth_abi_encode: Final[Callable[[TypeStrs, List[Any]], bytes]] = (
    eth_abi.encode if hasattr(eth_abi, "encode") else eth_abi.encode_abi
)
__eth_abi_decode: Final[Callable[[TypeStrs, hexbytes.HexBytes], Tuple[Any, ...]]] = (
    eth_abi.decode if hasattr(eth_abi, "decode") else eth_abi.decode_abi
)


def __encode_input(abi: AbiDict, signature: str, *args: Any) -> Union[HexStr, Exception]:
    try:
        data = format_input_but_cache_checksums(abi, args)
        types_list = get_type_strings(abi["inputs"])
        return signature + __eth_abi_encode(types_list, data).hex()  # type: ignore [return-value]
    except Exception as e:
        return e


_skip_proc_pool = {"0xcA11bde05977b3631167028862bE2a173976CA11"}  # multicall3
# NOTE: retry 429 errors if running multiple services on same rpc
while True:
    try:
        chainid = chain.id
        break
    except Exception as e:
        if "429" not in str(e):
            raise
if multicall2 := MULTICALL2_ADDRESSES.get(chainid, None):
    _skip_proc_pool.add(to_checksum_address(multicall2))


def __decode_output(hexstr: BytesLike, abi: AbiDict) -> Any:
    try:
        types_list = get_type_strings(abi["outputs"])
        result = __eth_abi_decode(types_list, HexBytes(hexstr))
        result = format_output_but_cache_checksums(abi, result)
        return result[0] if len(result) == 1 else result
    except DecodingError as e:
        e.args = *e.args, {"hexstr": hexstr, "abi": abi}
        return e
    except Exception as e:
        return e


def __validate_output(abi: AbiDict, hexstr: BytesLike) -> None:
    try:
        selector = HexBytes(hexstr)[:4].hex()
        if selector == "0x08c379a0":
            revert_str = eth_abi.decode_abi(["string"], HexBytes(hexstr)[4:])[0]
            raise Revert(f"Call reverted: {revert_str}")
        elif selector == "0x4e487b71":
            error_code = int(HexBytes(hexstr)[4:].hex(), 16)
            if error_code in SOLIDITY_ERROR_CODES:
                revert_str = SOLIDITY_ERROR_CODES[error_code]
            else:
                revert_str = f"Panic (error code: {error_code})"
            raise Revert(f"Call reverted: {revert_str}")
        elif selector == "0xc1b84b2f":
            raise Revert("Call reverted: execution reverted")
        if abi["outputs"] and not hexstr:
            raise Revert("No data was returned - the call likely reverted")
    except ValueError as e:
        try:
            raise VirtualMachineError(e) from None
        except:
            raise e from e.__cause__


# NOTE: We do a little monkey patching to save cpu cycles on checksums


def format_input_but_cache_checksums(abi: AbiDict, inputs: ListOrTuple[Any]) -> List[Any]:
    # Format contract inputs based on ABI types
    if len(inputs) and not len(abi["inputs"]):
        raise TypeError(f"{abi['name']} requires no arguments")
    abi_types = _get_abi_types(abi["inputs"])
    try:
        return _format_tuple_but_cache_checksums(abi_types, inputs)
    except Exception as e:
        raise type(e)(f"{abi['name']} {e}") from e


def format_output_but_cache_checksums(
    abi: AbiDict, outputs: ListOrTuple[Any]
) -> brownie.convert.datatypes.ReturnValue:
    # Format contract outputs based on ABI types
    abi_types = _get_abi_types(abi["outputs"])
    result = _format_tuple_but_cache_checksums(abi_types, outputs)
    return ReturnValue(result, abi["outputs"])


brownie.network.contract.format_input = format_input_but_cache_checksums
brownie.network.contract.format_output = format_output_but_cache_checksums


def _format_tuple_but_cache_checksums(
    abi_types: Sequence[ABIType], values: ListOrTuple[Any]
) -> List[Any]:
    result = []
    _check_array(values, len(abi_types))
    for type_, value in zip(abi_types, values):
        try:
            if type_.is_array:
                result.append(_format_array_but_cache_checksums(type_, value))
            elif isinstance(type_, TupleType):
                result.append(_format_tuple_but_cache_checksums(type_.components, value))
            else:
                result.append(_format_single_but_cache_checksums(type_.to_type_str(), value))
        except Exception as e:
            raise type(e)(f"'{value}' - {e}") from e
    return result


def _format_array_but_cache_checksums(abi_type: ABIType, values: ListOrTuple[Any]) -> List[Any]:
    _check_array(values, abi_type.arrlist[-1][0] if len(abi_type.arrlist[-1]) else None)
    item_type = abi_type.item_type
    if item_type.is_array:
        return [_format_array_but_cache_checksums(item_type, v) for v in values]
    elif isinstance(item_type, TupleType):
        components = item_type.components
        return [_format_tuple_but_cache_checksums(components, v) for v in values]
    else:
        type_str: str = item_type.to_type_str()
        return [_format_single_but_cache_checksums(type_str, v) for v in values]


def _format_single_but_cache_checksums(type_str: str, value: Any) -> Any:
    # Apply standard formatting to a single value
    if "uint" in type_str:
        return to_uint(value, type_str)
    elif "int" in type_str:
        return to_int(value, type_str)
    elif type_str == "fixed168x10":
        return to_decimal(value)
    elif type_str == "bool":
        return to_bool(value)
    elif type_str == "address":
        # NOTE: since we skip brownie's formatter we must ensure we pass in a usable type
        return to_checksum_address(
            value if isinstance(value, (str, bytes, bytearray, int, bool)) else str(value)
        )
    elif "byte" in type_str:
        return HexString(value, type_str)
    elif "string" in type_str:
        return to_string(value)
    raise TypeError(f"Unknown type: {type_str}")


# NOTE: The monkey patches above work for call input and output but the ones below help with event decoding.
brownie.convert.normalize._format_array = _format_array_but_cache_checksums
brownie.convert.normalize._format_single = _format_single_but_cache_checksums
brownie.convert.normalize._format_tuple = _format_tuple_but_cache_checksums
